"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test tc_tracks_synth module.
"""

import array
import unittest
from pathlib import Path

import numpy as np
import xarray as xr

import climada.hazard.tc_tracks as tc
import climada.hazard.tc_tracks_synth as tc_synth
import climada.util.coordinates
from climada.util.constants import TC_ANDREW_FL

DATA_DIR = Path(__file__).parent.joinpath('data')
TEST_TRACK = DATA_DIR.joinpath("trac_brb_test.csv")
TEST_TRACK_SHORT = DATA_DIR.joinpath("trac_short_test.csv")
TEST_TRACK_DECAY_END_OCEAN = DATA_DIR.joinpath('1997018S11059_gen3.nc')
TEST_TRACK_DECAY_END_OCEAN_HIST = DATA_DIR.joinpath('1997018S11059.nc')
TEST_TRACK_DECAY_PENV_GT_PCEN = DATA_DIR.joinpath('1988021S12080_gen2.nc')
TEST_TRACK_DECAY_PENV_GT_PCEN_HIST = DATA_DIR.joinpath('1988021S12080.nc')

class TestDecay(unittest.TestCase):
    def test_apply_decay_no_landfall_pass(self):
        """Test _apply_land_decay with no historical tracks with landfall"""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK_SHORT)
        extent = tc_track.get_extent()
        land_geom = climada.util.coordinates.get_land_geometry(
            extent=extent, resolution=10
        )
        tc.track_land_params(tc_track.data[0], land_geom)
        tc_track.data[0]['orig_event_flag'] = False
        tc_ref = tc_track.data[0].copy()
        tc_synth._apply_land_decay(tc_track.data, dict(), dict(), land_geom)

        self.assertTrue(np.allclose(tc_track.data[0].max_sustained_wind.values,
                                    tc_ref.max_sustained_wind.values))
        self.assertTrue(np.allclose(tc_track.data[0].central_pressure.values,
                                    tc_ref.central_pressure.values))
        self.assertTrue(np.allclose(tc_track.data[0].environmental_pressure.values,
                                    tc_ref.environmental_pressure.values))
        self.assertTrue(np.all(np.isnan(tc_track.data[0].dist_since_lf.values)))

    def test_apply_decay_pass(self):
        """Test _apply_land_decay against MATLAB reference."""
        v_rel = {
            4: 0.0038950967656296597,
            -1: 0.0038950967656296597,
            0: 0.0038950967656296597,
            1: 0.0038950967656296597,
            2: 0.0038950967656296597,
            3: 0.0038950967656296597,
            5: 0.0038950967656296597
        }

        p_rel = {
            4: (1.0499941, 0.007978940084158488),
            -1: (1.0499941, 0.007978940084158488),
            0: (1.0499941, 0.007978940084158488),
            1: (1.0499941, 0.007978940084158488),
            2: (1.0499941, 0.007978940084158488),
            3: (1.0499941, 0.007978940084158488),
            5: (1.0499941, 0.007978940084158488)
        }

        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TC_ANDREW_FL)
        tc_track.data[0]['orig_event_flag'] = False
        extent = tc_track.get_extent()
        land_geom = climada.util.coordinates.get_land_geometry(
            extent=extent, resolution=10
        )
        tc.track_land_params(tc_track.data[0], land_geom)
        tc_synth._apply_land_decay(tc_track.data, v_rel, p_rel, land_geom,
                                   s_rel=True, check_plot=False)

        p_ref = np.array([
            1.010000000000000, 1.009000000000000, 1.008000000000000,
            1.006000000000000, 1.003000000000000, 1.002000000000000,
            1.001000000000000, 1.000000000000000, 1.000000000000000,
            1.001000000000000, 1.002000000000000, 1.005000000000000,
            1.007000000000000, 1.010000000000000, 1.010000000000000,
            1.010000000000000, 1.010000000000000, 1.010000000000000,
            1.010000000000000, 1.007000000000000, 1.004000000000000,
            1.000000000000000, 0.994000000000000, 0.981000000000000,
            0.969000000000000, 0.961000000000000, 0.947000000000000,
            0.933000000000000, 0.922000000000000, 0.930000000000000,
            0.937000000000000, 0.951000000000000, 0.947000000000000,
            0.943000000000000, 0.948000000000000, 0.946000000000000,
            0.941000000000000, 0.937000000000000, 0.955000000000000,
            0.9741457117, 0.99244068917, 1.00086729492, 1.00545853355,
            1.00818354609, 1.00941850023, 1.00986192053, 1.00998400565
        ]) * 1e3

        self.assertTrue(np.allclose(p_ref, tc_track.data[0].central_pressure.values))

        v_ref = np.array([
            0.250000000000000, 0.300000000000000, 0.300000000000000,
            0.350000000000000, 0.350000000000000, 0.400000000000000,
            0.450000000000000, 0.450000000000000, 0.450000000000000,
            0.450000000000000, 0.450000000000000, 0.450000000000000,
            0.450000000000000, 0.400000000000000, 0.400000000000000,
            0.400000000000000, 0.400000000000000, 0.450000000000000,
            0.450000000000000, 0.500000000000000, 0.500000000000000,
            0.550000000000000, 0.650000000000000, 0.800000000000000,
            0.950000000000000, 1.100000000000000, 1.300000000000000,
            1.450000000000000, 1.500000000000000, 1.250000000000000,
            1.300000000000000, 1.150000000000000, 1.150000000000000,
            1.150000000000000, 1.150000000000000, 1.200000000000000,
            1.250000000000000, 1.250000000000000, 1.200000000000000,
            0.9737967353, 0.687255951, 0.4994850556, 0.3551480462, 0.2270548036,
            0.1302099557, 0.0645385918, 0.0225325851
        ]) * 1e2

        self.assertTrue(np.allclose(v_ref, tc_track.data[0].max_sustained_wind.values))

        cat_ref = tc.set_category(tc_track.data[0].max_sustained_wind.values,
                                  tc_track.data[0].max_sustained_wind_unit)
        self.assertEqual(cat_ref, tc_track.data[0].category)

    def test_func_decay_p_pass(self):
        """Test decay function for pressure with its inverse."""
        s_coef = 1.05
        b_coef = 0.04
        x_val = np.arange(0, 100, 10)
        res = tc_synth._decay_p_function(s_coef, b_coef, x_val)
        b_coef_res = tc_synth._solve_decay_p_function(s_coef, res, x_val)

        self.assertTrue(np.allclose(b_coef_res[1:], np.ones((x_val.size - 1,)) * b_coef))
        self.assertTrue(np.isnan(b_coef_res[0]))

    def test_func_decay_v_pass(self):
        """Test decay function for wind with its inverse."""
        a_coef = 0.04
        x_val = np.arange(0, 100, 10)
        res = tc_synth._decay_v_function(a_coef, x_val)
        a_coef_res = tc_synth._solve_decay_v_function(res, x_val)

        self.assertTrue(np.allclose(a_coef_res[1:], np.ones((x_val.size - 1,)) * a_coef))
        self.assertTrue(np.isnan(a_coef_res[0]))

    def test_decay_ps_value(self):
        """Test the calculation of S in pressure decay."""
        on_land_idx = 5
        tr_ds = xr.Dataset()
        tr_ds.coords['time'] = ('time', np.arange(10))
        tr_ds['central_pressure'] = ('time', np.arange(10, 20))
        tr_ds['environmental_pressure'] = ('time', np.arange(20, 30))
        tr_ds['on_land'] = ('time', np.zeros((10,)).astype(bool))
        tr_ds.on_land[on_land_idx] = True
        p_landfall = 100

        res = tc_synth._calc_decay_ps_value(tr_ds, p_landfall, on_land_idx, s_rel=True)
        self.assertEqual(res, float(tr_ds.environmental_pressure[on_land_idx] / p_landfall))
        res = tc_synth._calc_decay_ps_value(tr_ds, p_landfall, on_land_idx, s_rel=False)
        self.assertEqual(res, float(tr_ds.central_pressure[on_land_idx] / p_landfall))

    def test_calc_decay_no_landfall_pass(self):
        """Test _calc_land_decay with no historical tracks with landfall"""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK_SHORT)
        expected_warning = 'only %s historical tracks were provided. ' % len(tc_track.data)
        extent = tc_track.get_extent()
        land_geom = climada.util.coordinates.get_land_geometry(
            extent=extent, resolution=10
        )
        tc.track_land_params(tc_track.data[0], land_geom)
        with self.assertLogs('climada.hazard.tc_tracks_synth', level='INFO') as cm:
            tc_synth._calc_land_decay(tc_track.data, land_geom)
        self.assertIn(expected_warning, cm.output[0])
        self.assertIn('No historical track with landfall.', cm.output[1])

    def test_calc_land_decay_pass(self):
        """Test _calc_land_decay with environmental pressure function."""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TC_ANDREW_FL)
        extent = tc_track.get_extent()
        land_geom = climada.util.coordinates.get_land_geometry(
            extent=extent, resolution=10
        )
        tc.track_land_params(tc_track.data[0], land_geom)
        v_rel, p_rel = tc_synth._calc_land_decay(tc_track.data, land_geom)

        self.assertEqual(7, len(v_rel))
        for i, val in enumerate(v_rel.values()):
            self.assertAlmostEqual(val, 0.0038894834)
            self.assertTrue(i - 1 in v_rel.keys())

        self.assertEqual(7, len(p_rel))
        for i, val in enumerate(p_rel.values()):
            self.assertAlmostEqual(val[0], 1.0598491)
            self.assertAlmostEqual(val[1], 0.0041949237)
            self.assertTrue(i - 1 in p_rel.keys())

    def test_decay_values_andrew_pass(self):
        """Test _decay_values with central pressure function."""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TC_ANDREW_FL)
        s_rel = False
        extent = tc_track.get_extent()
        land_geom = climada.util.coordinates.get_land_geometry(
            extent=extent, resolution=10
        )
        tc.track_land_params(tc_track.data[0], land_geom)
        v_lf, p_lf, x_val = tc_synth._decay_values(tc_track.data[0], land_geom, s_rel)

        ss_category = 4
        s_cell_1 = 1 * [1.0149413347244263]
        s_cell_2 = 8 * [1.047120451927185]
        s_cell = s_cell_1 + s_cell_2
        p_vs_lf_time_relative = [
            1.0149413020277482, 1.018848167539267, 1.037696335078534,
            1.0418848167539267, 1.043979057591623, 1.0450261780104713,
            1.0460732984293193, 1.0471204188481675, 1.0471204188481675
        ]

        self.assertEqual(list(p_lf.keys()), [ss_category])
        self.assertEqual(p_lf[ss_category][0], array.array('f', s_cell))
        self.assertEqual(p_lf[ss_category][1], array.array('f', p_vs_lf_time_relative))

        v_vs_lf_time_relative = [
            0.8846153846153846, 0.6666666666666666, 0.4166666666666667,
            0.2916666666666667, 0.250000000000000, 0.250000000000000,
            0.20833333333333334, 0.16666666666666666, 0.16666666666666666
        ]
        self.assertEqual(list(v_lf.keys()), [ss_category])
        self.assertEqual(v_lf[ss_category], array.array('f', v_vs_lf_time_relative))

        x_val_ref = np.array([
            95.9512939453125, 53.624916076660156, 143.09530639648438,
            225.0262908935547, 312.5832824707031, 427.43109130859375,
            570.1857299804688, 750.3827514648438, 1020.5431518554688
        ])
        self.assertEqual(list(x_val.keys()), [ss_category])
        self.assertTrue(np.allclose(x_val[ss_category], x_val_ref))

    def test_decay_calc_coeff(self):
        """Test _decay_calc_coeff against MATLAB"""
        x_val = {
            4: np.array([
                53.57314960249573, 142.97903059281566, 224.76733726289183,
                312.14621544207563, 426.6757021862584, 568.9358305779094,
                748.3713215157885, 1016.9904230811956
            ])
        }

        v_lf = {
            4: np.array([
                0.6666666666666666, 0.4166666666666667, 0.2916666666666667,
                0.250000000000000, 0.250000000000000, 0.20833333333333334,
                0.16666666666666666, 0.16666666666666666
            ])
        }

        p_lf = {
            4: (8 * [1.0471204188481675],
                np.array([
                    1.018848167539267, 1.037696335078534, 1.0418848167539267,
                    1.043979057591623, 1.0450261780104713, 1.0460732984293193,
                    1.0471204188481675, 1.0471204188481675
                ])
            )
        }

        v_rel, p_rel = tc_synth._decay_calc_coeff(x_val, v_lf, p_lf)

        for i, val in enumerate(v_rel.values()):
            self.assertAlmostEqual(val, 0.004222091151737)
            self.assertTrue(i - 1 in v_rel.keys())

        for i, val in enumerate(p_rel.values()):
            self.assertAlmostEqual(val[0], 1.047120418848168)
            self.assertAlmostEqual(val[1], 0.008871782287614)
            self.assertTrue(i - 1 in v_rel.keys())

    def test_wrong_decay_pass(self):
        """Test decay not implemented when coefficient < 1"""
        track = tc.TCTracks.from_ibtracs_netcdf(provider='usa', storm_id='1975178N28281')

        track_gen = track.data[0]
        track_gen['lat'] = np.array([
            28.20340431, 28.7915261, 29.38642458, 29.97836984, 30.56844404,
            31.16265292, 31.74820301, 32.34449825, 32.92261894, 33.47430891,
            34.01492525, 34.56789399, 35.08810845, 35.55965893, 35.94835174,
            36.29355848, 36.45379561, 36.32473812, 36.07552209, 35.92224784,
            35.84144186, 35.78298537, 35.86090718, 36.02440372, 36.37555559,
            37.06207765, 37.73197352, 37.97524273, 38.05560287, 38.21901208,
            38.31486156, 38.30813367, 38.28481808, 38.28410366, 38.25894812,
            38.20583372, 38.22741099, 38.39970022, 38.68367797, 39.08329904,
            39.41434629, 39.424984, 39.31327716, 39.30336335, 39.31714429,
            39.27031932, 39.30848775, 39.48759833, 39.73326595, 39.96187967,
            40.26954226, 40.76882202, 41.40398607, 41.93809726, 42.60395785,
            43.57074792, 44.63816143, 45.61450458, 46.68528511, 47.89209365,
            49.15580502
        ])
        track_gen['lon'] = np.array([
            -79.20514075, -79.25243311, -79.28393082, -79.32324646,
            -79.36668585, -79.41495519, -79.45198688, -79.40580325,
            -79.34965443, -79.36938122, -79.30294825, -79.06809546,
            -78.70281969, -78.29418936, -77.82170609, -77.30034709,
            -76.79004969, -76.37038827, -75.98641014, -75.58383356,
            -75.18310414, -74.7974524, -74.3797645, -73.86393572, -73.37910948,
            -73.01059003, -72.77051313, -72.68011328, -72.66864779,
            -72.62579773, -72.56307717, -72.46607618, -72.35871353,
            -72.31120649, -72.15537583, -71.75577051, -71.25287498,
            -70.75527907, -70.34788946, -70.17518421, -70.04446577,
            -69.76582749, -69.44372386, -69.15881376, -68.84351922,
            -68.47890287, -68.04184565, -67.53541437, -66.94008642,
            -66.25596075, -65.53496635, -64.83491802, -64.12962685,
            -63.54118808, -62.72934383, -61.34915091, -59.72580755,
            -58.24404252, -56.71972992, -55.0809336, -53.31524758
        ])

        v_rel = {
            1: 0.002249541544102336,
            -1: 0.00046889526284203036,
            2: 0.002649273787364977,
            0: 0.0016426186150461349,
            3: 0.00246400811445618,
            5: 0.0030442198547309075,
            4: 0.002346537842810565,
        }
        p_rel = {
            1: (1.028420239620591, 0.003174733355067952),
            -1: (1.0046803184177564, 0.0007997633912500546),
            2: (1.0498749735343516, 0.0034665588904747515),
            0: (1.0140127424090262, 0.002131858515233042),
            3: (1.0619445995372885, 0.003467268426139696),
            5: (1.0894914184297835, 0.004315034379018768),
            4: (1.0714354641894077, 0.002783787561718677),
        }
        track_gen.attrs['orig_event_flag'] = False

        cp_ref = np.array([1012., 1012.])
        single_track = tc.TCTracks()
        single_track.data = [track_gen]
        extent = single_track.get_extent()
        land_geom = climada.util.coordinates.get_land_geometry(
            extent=extent, resolution=10
        )
        track_res = tc_synth._apply_decay_coeffs(track_gen, v_rel, p_rel, land_geom, True)
        self.assertTrue(np.array_equal(cp_ref, track_res.central_pressure[9:11]))

    def test_decay_end_ocean(self):
        """Test decay is applied after landfall if the track ends over the ocean"""
        # this track was generated without applying landfall decay
        # (i.e. with decay=False in tc_synth.calc_perturbed_trajectories)
        tracks_synth_nodecay_example = tc.TCTracks.from_netcdf(TEST_TRACK_DECAY_END_OCEAN)

        # apply landfall decay
        extent = tracks_synth_nodecay_example.get_extent()
        land_geom = climada.util.coordinates.get_land_geometry(
            extent=extent, resolution=10
        )
        tracks_synth_nodecay_example.data = tc_synth._apply_land_decay(
            tracks_synth_nodecay_example.data,
            tc_synth.LANDFALL_DECAY_V,
            tc_synth.LANDFALL_DECAY_P,
            land_geom)
        track = tracks_synth_nodecay_example.data[0]

        # read its corresponding historical track
        track_hist = tc.TCTracks.from_netcdf(TEST_TRACK_DECAY_END_OCEAN_HIST)
        track_hist = track_hist.data[0]

        # Part 1: is landfall applied after going back to the ocean?
        # get that last strip over the ocean
        lf_idx = tc._get_landfall_idx(track)
        last_lf_idx = lf_idx[-1][1]
        # only suitable if track ends over the ocean
        self.assertTrue(last_lf_idx < track.time.size-2,
                         'This test should be re-written, data not suitable')
        # check pressure and wind values
        p_hist_end = track_hist.central_pressure.values[last_lf_idx:]
        p_synth_end = track.central_pressure.values[last_lf_idx:]
        self.assertTrue(np.all(p_synth_end > p_hist_end))
        v_hist_end = track_hist.max_sustained_wind.values[last_lf_idx:]
        v_synth_end = track.max_sustained_wind.values[last_lf_idx:]
        self.assertTrue(np.all(v_synth_end < v_hist_end))

        # Part 2: is landfall applied in all landfalls?
        p_hist_lf = np.concatenate([track_hist.central_pressure.values[lfs:lfe]
                                    for lfs,lfe in zip(*lf_idx)])
        p_synth_lf = np.concatenate([track.central_pressure.values[lfs:lfe]
                                     for lfs,lfe in zip(*lf_idx)])
        v_hist_lf = np.concatenate([track_hist.max_sustained_wind.values[lfs:lfe]
                                    for lfs,lfe in zip(*lf_idx)])
        v_synth_lf = np.concatenate([track.max_sustained_wind.values[lfs:lfe]
                                     for lfs,lfe in zip(*lf_idx)])
        self.assertTrue(np.all(p_synth_lf > p_hist_lf))
        self.assertTrue(np.all(v_synth_lf < v_hist_lf))
        self.assertTrue(np.all(track.central_pressure.values <= track.environmental_pressure.values))

    def test_decay_penv_gt_pcen(self):
        """Test decay is applied if penv at end of landfall < pcen just before landfall"""
        # this track was generated without applying landfall decay
        # (i.e. with decay=False in tc_synth.calc_perturbed_trajectories)
        tracks_synth_nodecay_example = tc.TCTracks.from_netcdf(TEST_TRACK_DECAY_PENV_GT_PCEN)

        # apply landfall decay
        extent = tracks_synth_nodecay_example.get_extent()
        land_geom = climada.util.coordinates.get_land_geometry(
            extent=extent, resolution=10
        )
        tracks_synth_nodecay_example.data = tc_synth._apply_land_decay(
            tracks_synth_nodecay_example.data,
            tc_synth.LANDFALL_DECAY_V,
            tc_synth.LANDFALL_DECAY_P,
            land_geom)
        track = tracks_synth_nodecay_example.data[0]

        # read its corresponding historical track
        track_hist = tc.TCTracks.from_netcdf(TEST_TRACK_DECAY_PENV_GT_PCEN_HIST)
        track_hist = track_hist.data[0]

        # Part 1: is landfall applied after going back to the ocean?
        # get that last strip over the ocean
        lf_idx = tc._get_landfall_idx(track)
        start_lf_idx, end_lf_idx = lf_idx[0][0], lf_idx[1][0]

        # check pressure and wind values
        p_hist_end = track_hist.central_pressure.values[end_lf_idx:]
        p_synth_end = track.central_pressure.values[end_lf_idx:]
        self.assertTrue(np.all(p_synth_end > p_hist_end))
        v_hist_end = track_hist.max_sustained_wind.values[end_lf_idx:]
        v_synth_end = track.max_sustained_wind.values[end_lf_idx:]
        self.assertTrue(np.all(v_synth_end < v_hist_end))

        # Part 2: is landfall applied in all landfalls?

        # central pressure
        p_hist_lf = track_hist.central_pressure.values[start_lf_idx:end_lf_idx]
        p_synth_lf = track.central_pressure.values[start_lf_idx:end_lf_idx]
        # central pressure should be higher in synth than hist; unless it was set to p_env
        self.assertTrue(np.all(
            np.logical_or(p_synth_lf > p_hist_lf,
                          p_synth_lf == track.environmental_pressure.values[start_lf_idx:end_lf_idx])
            ))
        # but for this track is should be higher towards the end
        self.assertTrue(np.any(p_synth_lf > p_hist_lf))
        self.assertTrue(np.all(p_synth_lf >= p_hist_lf))

        # wind speed
        v_hist_lf = track_hist.max_sustained_wind.values[start_lf_idx:end_lf_idx]
        v_synth_lf = track.max_sustained_wind.values[start_lf_idx:end_lf_idx]
        # wind should decrease over time for that landfall
        v_before_lf = track_hist.max_sustained_wind.values[start_lf_idx-1]
        self.assertTrue(np.all(v_synth_lf[1:] < v_before_lf))
        # and wind speed should be lower in synth than hist at the end of and after this landfall
        self.assertTrue(np.all(
            track.max_sustained_wind.values[end_lf_idx:] < track_hist.max_sustained_wind.values[end_lf_idx:]
            ))
        # finally, central minus env pressure cannot increase during this landfall
        p_env_lf = track.central_pressure.values[start_lf_idx:end_lf_idx]
        self.assertTrue(np.all(np.diff(p_env_lf - p_synth_lf) <= 0))

class TestSynth(unittest.TestCase):
    def test_angle_funs_pass(self):
        """Test functions used by random walk code"""
        self.assertAlmostEqual(tc_synth._get_bearing_angle(np.array([15, 20]),
                                                           np.array([0, 0]))[0], 90.0)
        self.assertAlmostEqual(tc_synth._get_bearing_angle(np.array([20, 20]),
                                                           np.array([0, 5]))[0], 0.0)
        self.assertAlmostEqual(tc_synth._get_bearing_angle(np.array([0, 0.00001]),
                                                           np.array([0, 0.00001]))[0], 45)
        pt_north = tc_synth._get_destination_points(0, 0, 0, 1)
        self.assertAlmostEqual(pt_north[0], 0.0)
        self.assertAlmostEqual(pt_north[1], 1.0)
        pt_west = tc_synth._get_destination_points(0, 0, -90, 3)
        self.assertAlmostEqual(pt_west[0], -3.0)
        self.assertAlmostEqual(pt_west[1], 0.0)
        pt_test = tc_synth._get_destination_points(8.523224, 47.371102,
                                                   151.14161003, 52.80812463)
        self.assertAlmostEqual(pt_test[0], 31.144113)
        self.assertAlmostEqual(pt_test[1], -1.590347)

    def test_random_no_landfall_pass(self):
        """Test calc_perturbed_trajectories with decay and no historical tracks with landfall"""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK_SHORT)
        expected_warning = 'only %s historical tracks were provided. ' % len(tc_track.data)
        with self.assertLogs('climada.hazard.tc_tracks_synth', level='INFO') as cm:
            tc_track.calc_perturbed_trajectories(use_global_decay_params=False)
        self.assertIn(expected_warning, cm.output[1])
        self.assertIn('No historical track with landfall.', cm.output[2])

    def test_random_walk_ref_pass(self):
        """Test against MATLAB reference."""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK_SHORT)
        nb_synth_tracks = 2
        tc_track.calc_perturbed_trajectories(nb_synth_tracks=nb_synth_tracks, seed=25, decay=False)

        self.assertEqual(len(tc_track.data), nb_synth_tracks + 1)

        self.assertFalse(tc_track.data[1].orig_event_flag)
        self.assertEqual(tc_track.data[1].name, '1951239N12334_gen1')
        self.assertEqual(tc_track.data[1].id_no, 1.951239012334010e+12)
        self.assertAlmostEqual(tc_track.data[1].lon[0].values, -25.0448138)
        self.assertAlmostEqual(tc_track.data[1].lon[1].values, -25.74439739)
        self.assertAlmostEqual(tc_track.data[1].lon[2].values, -26.54491644)
        self.assertAlmostEqual(tc_track.data[1].lon[3].values, -27.73156829)
        self.assertAlmostEqual(tc_track.data[1].lon[4].values, -28.63175987)
        self.assertAlmostEqual(tc_track.data[1].lon[8].values, -34.05293373)

        self.assertAlmostEqual(tc_track.data[1].lat[0].values, 11.96825841)
        self.assertAlmostEqual(tc_track.data[1].lat[4].values, 11.86769405)
        self.assertAlmostEqual(tc_track.data[1].lat[5].values, 11.84378139)
        self.assertAlmostEqual(tc_track.data[1].lat[6].values, 11.85957282)
        self.assertAlmostEqual(tc_track.data[1].lat[7].values, 11.84555291)
        self.assertAlmostEqual(tc_track.data[1].lat[8].values, 11.8065998)

        self.assertFalse(tc_track.data[2].orig_event_flag)
        self.assertEqual(tc_track.data[2].name, '1951239N12334_gen2')
        self.assertAlmostEqual(tc_track.data[2].id_no, 1.951239012334020e+12)
        self.assertAlmostEqual(tc_track.data[2].lon[0].values, -25.47658461)
        self.assertAlmostEqual(tc_track.data[2].lon[3].values, -28.08465841)
        self.assertAlmostEqual(tc_track.data[2].lon[4].values, -28.85901852)
        self.assertAlmostEqual(tc_track.data[2].lon[8].values, -33.62144837)

        self.assertAlmostEqual(tc_track.data[2].lat[0].values, 11.82886685)
        self.assertAlmostEqual(tc_track.data[2].lat[6].values, 11.71068012)
        self.assertAlmostEqual(tc_track.data[2].lat[7].values, 11.69832976)
        self.assertAlmostEqual(tc_track.data[2].lat[8].values, 11.64145734)

    def test_random_walk_decay_pass(self):
        """Test land decay is called from calc_perturbed_trajectories."""
        assert TC_ANDREW_FL.is_file()
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TC_ANDREW_FL)
        nb_synth_tracks = 2
        # should work if using global parameters
        with self.assertLogs('climada.hazard.tc_tracks_synth', level='DEBUG') as cm0:
            tc_track.calc_perturbed_trajectories(nb_synth_tracks=nb_synth_tracks, seed=25, decay=True,
                                                 use_global_decay_params=True)
        self.assertEqual(len(cm0), 2)
        self.assertEqual(tc_track.size, 3)
        # but alert the user otherwise
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TC_ANDREW_FL)
        with self.assertLogs('climada.hazard.tc_tracks_synth', level='DEBUG') as cm:
            tc_track.calc_perturbed_trajectories(nb_synth_tracks=nb_synth_tracks, seed=25, decay=True,
                                                 use_global_decay_params=False)
        self.assertIn('No historical track of category Tropical Depression '
                      'with landfall.', cm.output[2])
        self.assertIn('Decay parameters from category Hurricane Cat. 4 taken.',
                      cm.output[3])
        self.assertIn('No historical track of category Hurricane Cat. 1 with '
                      'landfall.', cm.output[4])
        self.assertIn('Decay parameters from category Hurricane Cat. 4 taken.',
                      cm.output[5])
        self.assertIn('No historical track of category Hurricane Cat. 3 with '
                      'landfall. Decay parameters from category Hurricane Cat. '
                      '4 taken.', cm.output[6])
        self.assertIn('No historical track of category Hurricane Cat. 5 with '
                      'landfall.', cm.output[7])

    def test_random_walk_identical_pass(self):
        """Test 0 perturbation leads to identical tracks."""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TC_ANDREW_FL)
        nb_synth_tracks = 2
        tc_track.calc_perturbed_trajectories(nb_synth_tracks=nb_synth_tracks,
                                  max_shift_ini=0, max_dspeed_rel=0, max_ddirection=0, decay=False)
        orig_track = tc_track.data[0]
        for syn_track in tc_track.data[1:]:
            np.testing.assert_allclose(orig_track.lon.values, syn_track.lon.values, atol=1e-4)
            np.testing.assert_allclose(orig_track.lat.values, syn_track.lat.values, atol=1e-4)
            for varname in ["time", "time_step", "radius_max_wind", "max_sustained_wind",
                            "central_pressure", "environmental_pressure"]:
                np.testing.assert_array_equal(orig_track[varname].values,
                                              syn_track[varname].values)

    def test_random_walk_single_point(self):
        found = False
        for year in range(1951, 1981):
            tc_track = tc.TCTracks.from_ibtracs_netcdf(provider='usa',
                                         year_range=(year,year),
                                         discard_single_points=False)
            singlept = np.where([x.time.size == 1 for x in tc_track.data])[0]
            found = len(singlept) > 0
            if found:
                # found a case with a single-point track, keep max three tracks for efficiency
                tc_track.data = tc_track.data[max(0, singlept[0]-1):singlept[0]+2]
                n_tr = tc_track.size
                tc_track.equal_timestep()
                tc_track.calc_perturbed_trajectories(nb_synth_tracks=2)
                self.assertEqual((2+1)*n_tr, tc_track.size)
                break
        self.assertTrue(found)

    def test_cutoff_tracks(self):
        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id='1986226N30276')
        tc_track.equal_timestep()
        with self.assertLogs('climada.hazard.tc_tracks_synth', level='DEBUG') as cm:
            tc_track.calc_perturbed_trajectories(nb_synth_tracks=10)
        self.assertIn('The following generated synthetic tracks moved beyond '
                      'the range of [-70, 70] degrees latitude', cm.output[1])

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestDecay)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSynth))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Functionalities for the import and basic processing of ISIMIP output data (.nc)
Data to be downloaded from https://esg.pik-potsdam.de/search/isimip/
(requires ESG login, sript will be added later)


All functions should work for ISIMIP netcdf output files regardless of the resolution
of saptial (lat, lon) and temporal (time) resolution and dimension of input data.
Not that ISIMIP data comes in a range of resolutions, i.e. daily (e.g. discharge),
monthly, yearly (e.g. yield)
"""


import xarray as xr

bbox_world = [-85, 85, -180, 180]

def _read_one_nc(file_name, bbox=None, years=None):
    """Reads 1 ISIMIP output NETCDF file data within a certain bounding box and time period

    Parameters
    ----------
    file_name : str
        Absolute or relative path to *.nc
    bbox : array
        bounding box containing [Lon min, lat min, lon max, lat max]
    years : array
        start and end year of the time series that shall be extracted

    Returns
    -------
    data : dataset
        Contains data in the specified bounding box and for the
        specified time period
    """
    data = xr.open_dataset(file_name, decode_times=False)
    if not bbox:
        bbox = bbox_world
    if not years:
        return data.sel(lat=slice(bbox[3], bbox[1]), lon=slice(bbox[0], bbox[2]))

    time_id = years - int(data.time.units[12:16])
    return data.sel(lat=slice(bbox[3], bbox[1]), lon=slice(bbox[0], bbox[2]),
                    time=slice(time_id[0], time_id[1]))



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test Forecast class
"""

import unittest
import datetime as dt
import numpy as np
import geopandas as gpd
import matplotlib.pyplot as plt
import fiona
from cartopy.io import shapereader
from pathlib import Path

from climada import CONFIG
from climada.hazard.storm_europe import StormEurope
from climada.entity.exposures.base import Exposures, INDICATOR_IMPF
from climada.entity import ImpactFuncSet
from climada.entity.impact_funcs.storm_europe import ImpfStormEurope
from climada.engine.forecast import Forecast, FORECAST_PLOT_DIR
from climada.util.constants import WS_DEMO_NC

HAZ_DIR = CONFIG.hazard.test_data.dir()

class TestCalc(unittest.TestCase):
    """Test calc and propety functions from the Forecast class"""

    def test_Forecast_calc_properties(self):
        """Test calc and propety functions from the Forecast class"""
        #hazard
        haz = StormEurope.from_cosmoe_file(
            HAZ_DIR.joinpath('storm_europe_cosmoe_forecast_vmax_testfile.nc'),
            run_datetime=dt.datetime(2018,1,1),
            event_date=dt.datetime(2018,1,3))
        #exposure
        data = {}
        data['latitude'] = haz.centroids.lat
        data['longitude'] = haz.centroids.lon
        data['value'] = np.ones_like(data['latitude']) * 100000
        data['deductible'] = np.zeros_like(data['latitude'])
        data[INDICATOR_IMPF + 'WS'] = np.ones_like(data['latitude'])
        data['region_id'] = np.ones_like(data['latitude'],dtype=int) * 756
        expo = Exposures(gpd.GeoDataFrame(data=data))
        #vulnerability
        #generate vulnerability
        impact_function = ImpfStormEurope.from_welker()
        impact_function_set = ImpactFuncSet()
        impact_function_set.append(impact_function)
        #create and calculate Forecast
        forecast = Forecast({dt.datetime(2018,1,1): haz}, expo, impact_function_set)
        forecast.calc()
        # test
        self.assertEqual(len(forecast.run_datetime), 1)
        self.assertEqual(forecast.run_datetime[0], dt.datetime(2018,1,1))
        self.assertEqual(forecast.event_date, dt.datetime(2018,1,3))
        self.assertEqual(forecast.lead_time().days,2)
        self.assertEqual(forecast.summary_str(),
                         'WS_NWP_run2018010100_event20180103_Switzerland')
        self.assertAlmostEqual(forecast.ai_agg(), 26.347, places=1)
        self.assertAlmostEqual(forecast.ei_exp()[1], 7.941, places=1)
        self.assertEqual(len(forecast.hazard), 1)
        self.assertIsInstance(forecast.hazard[0], StormEurope)
        self.assertIsInstance(forecast.exposure, Exposures)
        self.assertIsInstance(forecast.vulnerability, ImpactFuncSet)

    def test_Forecast_init_raise(self):
        """Test calc and propety functions from the Forecast class"""
        #hazard with several event dates
        storms = StormEurope.from_footprints(WS_DEMO_NC, description='test_description')
        #exposure
        data = {}
        data['latitude'] = np.array([1, 2, 3])
        data['longitude'] = np.array([1, 2, 3])
        data['value'] = np.ones_like(data['latitude']) * 100000
        data['deductible'] = np.zeros_like(data['latitude'])
        data[INDICATOR_IMPF + 'WS'] = np.ones_like(data['latitude'])
        data['region_id'] = np.ones_like(data['latitude'],dtype=int) * 756
        expo = Exposures(gpd.GeoDataFrame(data=data))
        #vulnerability
        #generate vulnerability
        impact_function_set = ImpactFuncSet()
        #create and calculate Forecast
        with self.assertRaises(ValueError):
            Forecast({dt.datetime(2018,1,1): storms}, expo, impact_function_set)


class TestPlot(unittest.TestCase):
    """Test plotting functions from the Forecast class"""

    def test_Forecast_plot(self):
        """Test cplotting functions from the Forecast class"""
                #hazard
        haz1 = StormEurope.from_cosmoe_file(
            HAZ_DIR.joinpath('storm_europe_cosmoe_forecast_vmax_testfile.nc'),
            run_datetime=dt.datetime(2018,1,1),
            event_date=dt.datetime(2018,1,3))
        haz1.centroids.lat += 0.6
        haz1.centroids.lon -= 1.2
        haz2 = StormEurope.from_cosmoe_file(
            HAZ_DIR.joinpath('storm_europe_cosmoe_forecast_vmax_testfile.nc'),
            run_datetime=dt.datetime(2018,1,1),
            event_date=dt.datetime(2018,1,3))
        haz2.centroids.lat += 0.6
        haz2.centroids.lon -= 1.2
        #exposure
        data = {}
        data['latitude'] = haz1.centroids.lat
        data['longitude'] = haz1.centroids.lon
        data['value'] = np.ones_like(data['latitude']) * 100000
        data['deductible'] = np.zeros_like(data['latitude'])
        data[INDICATOR_IMPF + 'WS'] = np.ones_like(data['latitude'])
        data['region_id'] = np.ones_like(data['latitude'],dtype=int) * 756
        expo = Exposures(gpd.GeoDataFrame(data=data))
        #vulnerability
        #generate vulnerability
        impact_function = ImpfStormEurope.from_welker()
        impact_function_set = ImpactFuncSet()
        impact_function_set.append(impact_function)
        #create and calculate Forecast
        forecast = Forecast({dt.datetime(2018,1,2): haz1,
                             dt.datetime(2017,12,31): haz2},
                            expo,
                            impact_function_set)
        forecast.calc()
        #create a file containing the polygons of Swiss cantons using natural earth
        cantons_file = CONFIG.local_data.save_dir.dir() / 'CHE_cantons.shp'
        adm1_shape_file = shapereader.natural_earth(resolution='10m',
                                                    category='cultural',
                                                    name='admin_1_states_provinces')
        if not cantons_file.exists():
            with fiona.open(adm1_shape_file, 'r') as source:
                with fiona.open(
                        cantons_file, 'w',
                        **source.meta) as sink:
                    for f in source:
                        if f['properties']['adm0_a3'] == 'CHE':
                            sink.write(f)
        #test plotting functions
        forecast.plot_imp_map(run_datetime=dt.datetime(2017,12,31),
                              polygon_file=str(cantons_file),
                              save_fig=True, close_fig=True)
        map_file_name = (forecast.summary_str(dt.datetime(2017,12,31)) +
                         '_impact_map' +
                         '.jpeg')
        map_file_name_full = Path(FORECAST_PLOT_DIR) / map_file_name
        map_file_name_full.absolute().unlink(missing_ok=False)
        forecast.plot_hist(run_datetime=dt.datetime(2017,12,31),
                           save_fig=False, close_fig=True)
        forecast.plot_exceedence_prob(run_datetime=dt.datetime(2017,12,31),
                                      threshold=5000, save_fig=False, close_fig=True)


        forecast.plot_warn_map(str(cantons_file),
                               decision_level = 'polygon',
                               thresholds=[100000,500000,
                                           1000000,5000000],
                               probability_aggregation='mean',
                               area_aggregation='sum',
                               title="Building damage warning",
                               explain_text="warn level based on aggregated damages",
                               save_fig=False,
                               close_fig=True)
        forecast.plot_warn_map(str(cantons_file),
                               decision_level = 'exposure_point',
                               thresholds=[1,1000,
                                           5000,5000000],
                               probability_aggregation=0.2,
                               area_aggregation=0.2,
                               title="Building damage warning",
                               explain_text="warn level based on aggregated damages",
                               run_datetime=dt.datetime(2017,12,31),
                               save_fig=False,
                               close_fig=True)
        forecast.plot_hexbin_ei_exposure()
        plt.close()
        with self.assertRaises(ValueError):
            forecast.plot_warn_map(str(cantons_file),
                                   decision_level = 'test_fail',
                                   probability_aggregation=0.2,
                                   area_aggregation=0.2,
                                   title="Building damage warning",
                                   explain_text="warn level based on aggregated damages",
                                   save_fig=False,
                                   close_fig=True)
        plt.close()
        with self.assertRaises(ValueError):
            forecast.plot_warn_map(str(cantons_file),
                                   decision_level = 'exposure_point',
                                   probability_aggregation='test_fail',
                                   area_aggregation=0.2,
                                   title="Building damage warning",
                                   explain_text="warn level based on aggregated damages",
                                   save_fig=False,
                                   close_fig=True)
        plt.close()
        with self.assertRaises(ValueError):
            forecast.plot_warn_map(str(cantons_file),
                                   decision_level = 'exposure_point',
                                   probability_aggregation=0.2,
                                   area_aggregation='test_fail',
                                   title="Building damage warning",
                                   explain_text="warn level based on aggregated damages",
                                   save_fig=False,
                                   close_fig=True)
        plt.close()


# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestCalc)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPlot))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test ImpactFunc class.
"""

import unittest
import numpy as np

from climada.entity.impact_funcs.base import ImpactFunc

class TestInterpolation(unittest.TestCase):
    """Impact function interpolation test"""

    def test_calc_mdr_pass(self):
        """Compute mdr interpolating values."""
        imp_fun = ImpactFunc()
        imp_fun.intensity = np.arange(0, 100, 10)
        imp_fun.paa = np.arange(0, 1, 0.1)
        imp_fun.mdd = np.arange(0, 1, 0.1)
        new_inten = 17.2
        self.assertEqual(imp_fun.calc_mdr(new_inten), 0.029583999999999996)

    def test_set_step(self):
        """Check default impact function: step function"""
        inten = (0, 5, 10)
        imp_fun = ImpactFunc.from_step_impf(inten)
        self.assertTrue(np.array_equal(imp_fun.paa, np.ones(4)))
        self.assertTrue(np.array_equal(imp_fun.mdd, np.array([0, 0, 1, 1])))
        self.assertTrue(np.array_equal(imp_fun.intensity, np.array([0, 5, 5, 10])))

    def test_set_sigmoid(self):
        """Check default impact function: sigmoid function"""
        inten = (0, 100, 5)
        imp_fun = ImpactFunc.from_sigmoid_impf(inten, L=1.0, k=2., x0=50.)
        self.assertTrue(np.array_equal(imp_fun.paa, np.ones(20)))
        self.assertEqual(imp_fun.mdd[10], 0.5)
        self.assertEqual(imp_fun.mdd[-1], 1.0)
        self.assertTrue(np.array_equal(imp_fun.intensity, np.arange(0, 100, 5)))

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestInterpolation)
    unittest.TextTestRunner(verbosity=2).run(TESTS)






"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Regroup methods to obtain images from Google Earth Engine API
"""

import logging
import webbrowser

# This module works only if you have a Google Earth Engine account.
# That's why `earthengine-api` is not in the CLIMADA requirements.
# See tutorial: climada_util_earth_engine.ipynb
# pylint: disable=import-error
import ee

LOGGER = logging.getLogger(__name__)
ee.Initialize()


def obtain_image_landsat_composite(landsat_collection, time_range, area):
    """Selection of Landsat cloud-free composites in the Earth Engine library
    See also: https://developers.google.com/earth-engine/landsat

    Parameters
    ----------
    collection :
        name of the collection
    time_range : ['YYYY-MT-DY','YYYY-MT-DY']
        must be inside the available data
    area : ee.geometry.Geometry
        area of interest

    Returns
    -------
    image_composite : ee.image.Image
     """
    collection = ee.ImageCollection(landsat_collection)

    # Filter by time range and location
    collection_time = collection.filterDate(time_range[0], time_range[1])
    image_area = collection_time.filterBounds(area)
    image_composite = ee.Algorithms.Landsat.simpleComposite(image_area, 75, 3)
    return image_composite

def obtain_image_median(collection, time_range, area):
    """Selection of median from a collection of images in the Earth Engine library
    See also: https://developers.google.com/earth-engine/reducers_image_collection

    Parameters
    ----------
    collection :
        name of the collection
    time_range : ['YYYY-MT-DY','YYYY-MT-DY']
        must be inside the available data
    area : ee.geometry.Geometry
        area of interest

    Returns
    -------
    image_median : ee.image.Image
     """
    collection = ee.ImageCollection(collection)

    # Filter by time range and location
    collection_time = collection.filterDate(time_range[0], time_range[1])
    image_area = collection_time.filterBounds(area)
    image_median = image_area.median()
    return image_median

def obtain_image_sentinel(sentinel_collection, time_range, area):
    """Selection of median, cloud-free image from a collection of images in the Sentinel 2 dataset
    See also: https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_S2

    Parameters
    ----------
    collection :
        name of the collection
    time_range : ['YYYY-MT-DY','YYYY-MT-DY']
        must be inside the available data
    area : ee.geometry.Geometry
        area of interest

    Returns
    -------
    sentinel_median : ee.image.Image
     """
# First, method to remove cloud from the image
    def maskclouds(image):
        band_qa = image.select('QA60')
        cloud_mask = ee.Number(2).pow(10).int()
        cirrus_mask = ee.Number(2).pow(11).int()
        mask = band_qa.bitwiseAnd(cloud_mask).eq(0) and (band_qa.bitwiseAnd(cirrus_mask).eq(0))
        return image.updateMask(mask).divide(10000)

    sentinel_filtered = (ee.ImageCollection(sentinel_collection).
                         filterBounds(area).
                         filterDate(time_range[0], time_range[1]).
                         filter(ee.Filter.lt('CLOUDY_PIXEL_PERCENTAGE', 20)).
                         map(maskclouds))

    sentinel_median = sentinel_filtered.median()
    return sentinel_median


def get_region(geom):
    """Get the region of a given geometry, needed for exporting tasks.

    Parameters
    ----------
    geom : ee.Geometry, ee.Feature, ee.Image
        region of interest

    Returns
    -------
    region : list
    """
    if isinstance(geom, ee.Geometry):
        region = geom.getInfo()["coordinates"]
    elif isinstance(geom, ee.Feature, ee.Image):
        region = geom.geometry().getInfo()["coordinates"]
    elif isinstance(geom, list):
        condition = all([isinstance(item) == list for item in geom])
        if condition:
            region = geom
    return region

def get_url(name, image, scale, region):
    """It will open and download automatically a zip folder containing Geotiff data of 'image'.
    If additional parameters are needed, see also:
    https://github.com/google/earthengine-api/blob/master/python/ee/image.py

    Parameters
    ----------
    name : str
        name of the created folder
    image : ee.image.Image
        image to export
    scale : int
        resolution of export in meters (e.g: 30 for Landsat)
    region : list
        region of interest

    Returns
    -------
    path : str
     """
    path = image.getDownloadURL({
        'name': (name),
        'scale': scale,
        'region': (region)
    })

    webbrowser.open_new_tab(path)
    return path



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test impf_TropCycl class.
"""

import unittest
import numpy as np
import pandas as pd

from climada.entity.impact_funcs.trop_cyclone import ImpfTropCyclone
from climada.entity.impact_funcs.trop_cyclone import ImpfSetTropCyclone

class TestEmanuelFormula(unittest.TestCase):
    """Impact function interpolation test"""

    def test_default_values_pass(self):
        """Compute mdr interpolating values."""
        imp_fun = ImpfTropCyclone.from_emanuel_usa()
        self.assertEqual(imp_fun.name, 'Emanuel 2011')
        self.assertEqual(imp_fun.haz_type, 'TC')
        self.assertEqual(imp_fun.id, 1)
        self.assertEqual(imp_fun.intensity_unit, 'm/s')
        self.assertTrue(np.array_equal(imp_fun.intensity, np.arange(0, 121, 5)))
        self.assertTrue(np.array_equal(imp_fun.paa, np.ones((25,))))
        self.assertTrue(np.array_equal(imp_fun.mdd[0:6], np.zeros((6,))))
        self.assertTrue(np.array_equal(imp_fun.mdd[6:10],
                        np.array([0.0006753419543492556, 0.006790495604105169,
                                  0.02425254393374475, 0.05758706257339458])))
        self.assertTrue(np.array_equal(imp_fun.mdd[10:15],
                        np.array([0.10870556455111065, 0.1761433569521351,
                                  0.2553983618763961, 0.34033822528795565,
                                  0.4249447743109498])))
        self.assertTrue(np.array_equal(imp_fun.mdd[15:20],
                        np.array([0.5045777092933046, 0.576424302849412,
                                  0.6393091739184916, 0.6932203123193963,
                                  0.7388256596555696])))
        self.assertTrue(np.array_equal(imp_fun.mdd[20:25],
                        np.array([0.777104531116526, 0.8091124649261859,
                                  0.8358522190681132, 0.8582150905529946,
                                  0.8769633232141456])))

    def test_values_pass(self):
        """Compute mdr interpolating values."""
        imp_fun = ImpfTropCyclone.from_emanuel_usa(impf_id=5,
                                                   intensity=np.arange(0, 6, 1),
                                                   v_thresh=2,
                                                   v_half=5,
                                                   scale=0.5)
        self.assertEqual(imp_fun.name, 'Emanuel 2011')
        self.assertEqual(imp_fun.haz_type, 'TC')
        self.assertEqual(imp_fun.id, 5)
        self.assertEqual(imp_fun.intensity_unit, 'm/s')
        self.assertTrue(np.array_equal(imp_fun.intensity, np.arange(0, 6, 1)))
        self.assertTrue(np.array_equal(imp_fun.paa, np.ones((6,))))
        self.assertTrue(np.array_equal(imp_fun.mdd[0:3], np.zeros((3,))))
        self.assertTrue(np.array_equal(imp_fun.mdd[3:],
                        np.array([0.017857142857142853, 0.11428571428571425,
                                  0.250000000000000])))

    def test_wrong_shape(self):
        """Set shape parameters."""
        with self.assertRaises(ValueError):
            imp_fun = ImpfTropCyclone.from_emanuel_usa(impf_id=5, v_thresh=2,
                                                       v_half=1,
                                                       intensity=np.arange(0, 6, 1))

    def test_wrong_scale(self):
        """Set shape parameters."""
        with self.assertRaises(ValueError):
            imp_fun = ImpfTropCyclone.from_emanuel_usa(impf_id=5, scale=2,
                                                       intensity=np.arange(0, 6, 1))

class TestCalibratedImpfSet(unittest.TestCase):
    """Test inititation of IFS with regional calibrated TC IFs
        based on Eberenz et al. (2020)"""

    def test_default_values_pass(self):
        """Test return TDR optimized IFs (TDR=1)"""
        impfs = ImpfSetTropCyclone.from_calibrated_regional_ImpfSet()
        v_halfs = ImpfSetTropCyclone.calibrated_regional_vhalf()
        # extract IF for region WP4
        impf_wp4 = impfs.get_func(fun_id=9)[0]
        self.assertIn('TC', impfs.get_ids().keys())
        self.assertEqual(impfs.size(), 10)
        self.assertEqual(impfs.get_ids()['TC'], [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.assertEqual(impf_wp4.intensity_unit, 'm/s')
        self.assertEqual(impf_wp4.name, 'North West Pacific (WP4)')
        self.assertAlmostEqual(v_halfs['WP2'], 188.4, places=7)
        self.assertAlmostEqual(v_halfs['ROW'], 110.1, places=7)
        self.assertListEqual(list(impf_wp4.intensity), list(np.arange(0, 121, 5)))
        self.assertEqual(impf_wp4.paa.min(), 1.)
        self.assertEqual(impf_wp4.mdd.min(), 0.0)
        self.assertAlmostEqual(impf_wp4.mdd.max(), 0.15779133833203, places=5)
        self.assertAlmostEqual(impf_wp4.calc_mdr(75), 0.02607326527808, places=5)

    def test_RMSF_pass(self):
        """Test return RMSF optimized impact function set (RMSF=minimum)"""
        impfs = ImpfSetTropCyclone.from_calibrated_regional_ImpfSet('RMSF')
        v_halfs = ImpfSetTropCyclone.calibrated_regional_vhalf(calibration_approach='RMSF')
        # extract IF for region NA1
        impf_na1 = impfs.get_func(fun_id=1)[0]
        self.assertEqual(impfs.size(), 10)
        self.assertEqual(impfs.get_ids()['TC'], [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.assertEqual(impf_na1.intensity_unit, 'm/s')
        self.assertEqual(impf_na1.name, 'Caribbean and Mexico (NA1)')
        self.assertAlmostEqual(v_halfs['NA1'], 59.6, places=7)
        self.assertAlmostEqual(v_halfs['ROW'], 73.4, places=7)
        self.assertListEqual(list(impf_na1.intensity), list(np.arange(0, 121, 5)))
        self.assertEqual(impf_na1.mdd.min(), 0.0)
        self.assertAlmostEqual(impf_na1.mdd.max(), 0.95560418241669, places=5)
        self.assertAlmostEqual(impf_na1.calc_mdr(75), 0.7546423895457, places=5)

    def test_quantile_pass(self):
        """Test return impact function set from quantile of inidividual event fitting (EDR=1)"""
        impfs = ImpfSetTropCyclone.from_calibrated_regional_ImpfSet('EDR')
        impfs_p10 = ImpfSetTropCyclone.from_calibrated_regional_ImpfSet('EDR', q=.1)
        # extract IF for region SI
        impf_si = impfs.get_func(fun_id=5)[0]
        impf_si_p10 = impfs_p10.get_func(fun_id=5)[0]
        self.assertEqual(impfs.size(), 10)
        self.assertEqual(impfs_p10.size(), 10)
        self.assertEqual(impf_si.intensity_unit, 'm/s')
        self.assertEqual(impf_si_p10.name, 'South Indian (SI)')
        self.assertAlmostEqual(impf_si_p10.mdd.max(), 0.99999999880, places=5)
        self.assertAlmostEqual(impf_si.calc_mdr(30), 0.01620503041, places=5)
        intensity = np.random.randint(26, impf_si.intensity.max())
        self.assertTrue(impf_si.calc_mdr(intensity) < impf_si_p10.calc_mdr(intensity))

    def test_get_countries_per_region(self):
        """Test static get_countries_per_region()"""
        ifs = ImpfSetTropCyclone()
        out = ifs.get_countries_per_region('NA2')
        self.assertEqual(out[0], 'USA and Canada')
        self.assertEqual(out[1], 2)
        self.assertListEqual(out[2], [124, 840])
        self.assertListEqual(out[3], ['CAN', 'USA'])

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestEmanuelFormula)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestCalibratedImpfSet))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test Impact class.
"""
import unittest
from pathlib import Path
import numpy as np
from scipy import sparse

from climada.entity.tag import Tag
from climada.hazard.tag import Tag as TagHaz
from climada.entity.entity_def import Entity
from climada.hazard.base import Hazard
from climada.engine.impact import Impact
from climada.util.constants import ENT_DEMO_TODAY, DEF_CRS, DEMO_DIR
from climada.util.api_client import Client
import climada.util.coordinates as u_coord
import climada.engine.test as engine_test


def get_haz_test_file(ds_name):
    client = Client()
    test_ds = client.get_dataset_info(name=ds_name, status='test_dataset')
    _, [haz_test_file] = client.download_dataset(test_ds)
    return haz_test_file


HAZ_TEST_MAT = get_haz_test_file('atl_prob_no_name')

DATA_FOLDER = DEMO_DIR / 'test-results'
DATA_FOLDER.mkdir(exist_ok=True)


class TestFreqCurve(unittest.TestCase):
    """Test exceedence frequency curve computation"""
    def test_ref_value_pass(self):
        """Test result against reference value"""
        imp = Impact()
        imp.frequency = np.ones(10) * 6.211180124223603e-04
        imp.at_event = np.zeros(10)
        imp.at_event[0] = 0
        imp.at_event[1] = 0.400665463736549e9
        imp.at_event[2] = 3.150330960044466e9
        imp.at_event[3] = 3.715826406781887e9
        imp.at_event[4] = 2.900244271902339e9
        imp.at_event[5] = 0.778570745161971e9
        imp.at_event[6] = 0.698736262566472e9
        imp.at_event[7] = 0.381063674256423e9
        imp.at_event[8] = 0.569142464157450e9
        imp.at_event[9] = 0.467572545849132e9
        imp.unit = 'USD'

        ifc = imp.calc_freq_curve()
        self.assertEqual(10, len(ifc.return_per))
        self.assertEqual(1610.0000000000000, ifc.return_per[9])
        self.assertEqual(805.00000000000000, ifc.return_per[8])
        self.assertEqual(536.66666666666663, ifc.return_per[7])
        self.assertEqual(402.500000000000, ifc.return_per[6])
        self.assertEqual(322.000000000000, ifc.return_per[5])
        self.assertEqual(268.33333333333331, ifc.return_per[4])
        self.assertEqual(230.000000000000, ifc.return_per[3])
        self.assertEqual(201.250000000000, ifc.return_per[2])
        self.assertEqual(178.88888888888889, ifc.return_per[1])
        self.assertEqual(161.000000000000, ifc.return_per[0])
        self.assertEqual(10, len(ifc.impact))
        self.assertEqual(3.715826406781887e9, ifc.impact[9])
        self.assertEqual(3.150330960044466e9, ifc.impact[8])
        self.assertEqual(2.900244271902339e9, ifc.impact[7])
        self.assertEqual(0.778570745161971e9, ifc.impact[6])
        self.assertEqual(0.698736262566472e9, ifc.impact[5])
        self.assertEqual(0.569142464157450e9, ifc.impact[4])
        self.assertEqual(0.467572545849132e9, ifc.impact[3])
        self.assertEqual(0.400665463736549e9, ifc.impact[2])
        self.assertEqual(0.381063674256423e9, ifc.impact[1])
        self.assertEqual(0, ifc.impact[0])
        self.assertEqual('Exceedance frequency curve', ifc.label)
        self.assertEqual('USD', ifc.unit)

    def test_ref_value_rp_pass(self):
        """Test result against reference value with given return periods"""
        imp = Impact()
        imp.frequency = np.ones(10) * 6.211180124223603e-04
        imp.at_event = np.zeros(10)
        imp.at_event[0] = 0
        imp.at_event[1] = 0.400665463736549e9
        imp.at_event[2] = 3.150330960044466e9
        imp.at_event[3] = 3.715826406781887e9
        imp.at_event[4] = 2.900244271902339e9
        imp.at_event[5] = 0.778570745161971e9
        imp.at_event[6] = 0.698736262566472e9
        imp.at_event[7] = 0.381063674256423e9
        imp.at_event[8] = 0.569142464157450e9
        imp.at_event[9] = 0.467572545849132e9
        imp.unit = 'USD'

        ifc = imp.calc_freq_curve(np.array([100, 500, 1000]))
        self.assertEqual(3, len(ifc.return_per))
        self.assertEqual(100, ifc.return_per[0])
        self.assertEqual(500, ifc.return_per[1])
        self.assertEqual(1000, ifc.return_per[2])
        self.assertEqual(3, len(ifc.impact))
        self.assertEqual(0, ifc.impact[0])
        self.assertEqual(2320408028.5695677, ifc.impact[1])
        self.assertEqual(3287314329.129928, ifc.impact[2])
        self.assertEqual('Exceedance frequency curve', ifc.label)
        self.assertEqual('USD', ifc.unit)

class TestOneExposure(unittest.TestCase):
    """Test one_exposure function"""
    def test_ref_value_insure_pass(self):
        """Test result against reference value"""
        # Read demo entity values
        # Set the entity default file to the demo one
        ent = Entity.from_excel(ENT_DEMO_TODAY)
        ent.check()

        # Read default hazard file
        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        # Create impact object
        impact = Impact()
        impact.at_event = np.zeros(hazard.intensity.shape[0])
        impact.eai_exp = np.zeros(len(ent.exposures.gdf.value))
        impact.tot_value = 0

        # Assign centroids to exposures
        ent.exposures.assign_centroids(hazard)

        # Compute impact for 6th exposure
        iexp = 5
        # Take its impact function
        imp_id = ent.exposures.gdf.impf_TC[iexp]
        imp_fun = ent.impact_funcs.get_func(hazard.tag.haz_type, imp_id)
        # Compute
        insure_flag = True
        impact._exp_impact(np.array([iexp]), ent.exposures, hazard, imp_fun, insure_flag)

        self.assertEqual(impact.eai_exp.size, ent.exposures.gdf.shape[0])
        self.assertEqual(impact.at_event.size, hazard.intensity.shape[0])

        events_pos = hazard.intensity[:, ent.exposures.gdf.centr_TC[iexp]].nonzero()[0]
        res_exp = np.zeros((ent.exposures.gdf.shape[0]))
        res_exp[iexp] = np.sum(impact.at_event[events_pos] * hazard.frequency[events_pos])
        np.testing.assert_array_equal(res_exp, impact.eai_exp)

        self.assertEqual(0, impact.at_event[12])
        # Check first 3 values
        self.assertEqual(0, impact.at_event[12])
        self.assertEqual(0, impact.at_event[41])
        self.assertEqual(1.0626600695059455e+06, impact.at_event[44])

        # Check intermediate values
        self.assertEqual(0, impact.at_event[6281])
        self.assertEqual(0, impact.at_event[4998])
        self.assertEqual(0, impact.at_event[9527])
        self.assertEqual(1.3318063850487845e+08, impact.at_event[7192])
        self.assertEqual(4.667108555054083e+06, impact.at_event[8624])

        # Check last 3 values
        self.assertEqual(0, impact.at_event[14349])
        self.assertEqual(0, impact.at_event[14347])
        self.assertEqual(0, impact.at_event[14309])

class TestCalc(unittest.TestCase):
    """Test impact calc method."""

    def test_ref_value_pass(self):
        """Test result against reference value"""
        # Read default entity values
        ent = Entity.from_excel(ENT_DEMO_TODAY)
        ent.check()

        # Read default hazard file
        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        # Create impact object
        impact = Impact()

        # Assign centroids to exposures
        ent.exposures.assign_centroids(hazard)

        # Compute the impact over the whole exposures
        impact.calc(ent.exposures, ent.impact_funcs, hazard)

        # Check result
        num_events = len(hazard.event_id)
        num_exp = ent.exposures.gdf.shape[0]
        # Check relative errors as well when absolute value gt 1.0e-7
        # impact.at_event == EDS.damage in MATLAB
        self.assertEqual(num_events, len(impact.at_event))
        self.assertEqual(0, impact.at_event[0])
        self.assertEqual(0, impact.at_event[int(num_events / 2)])
        self.assertAlmostEqual(1.472482938320243e+08, impact.at_event[13809])
        self.assertEqual(7.076504723057620e+10, impact.at_event[12147])
        self.assertEqual(0, impact.at_event[num_events - 1])
        # impact.eai_exp == EDS.ED_at_centroid in MATLAB
        self.assertEqual(num_exp, len(impact.eai_exp))
        self.assertAlmostEqual(1.518553670803242e+08, impact.eai_exp[0])
        self.assertAlmostEqual(1.373490457046383e+08, impact.eai_exp[int(num_exp / 2)], 6)
        self.assertAlmostEqual(1.373490457046383e+08, impact.eai_exp[int(num_exp / 2)], 5)
        self.assertAlmostEqual(1.066837260150042e+08, impact.eai_exp[num_exp - 1], 6)
        self.assertAlmostEqual(1.066837260150042e+08, impact.eai_exp[int(num_exp - 1)], 5)
        # impact.tot_value == EDS.Value in MATLAB
        # impact.aai_agg == EDS.ED in MATLAB
        self.assertAlmostEqual(6.570532945599105e+11, impact.tot_value)
        self.assertAlmostEqual(6.512201157564421e+09, impact.aai_agg, 5)
        self.assertAlmostEqual(6.512201157564421e+09, impact.aai_agg, 5)

    def test_calc_imp_mat_pass(self):
        """Test save imp_mat"""
        # Read default entity values
        ent = Entity.from_excel(ENT_DEMO_TODAY)
        ent.check()

        # Read default hazard file
        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        # Create impact object
        impact = Impact()

        # Assign centroids to exposures
        ent.exposures.assign_centroids(hazard)

        # Compute the impact over the whole exposures
        impact.calc(ent.exposures, ent.impact_funcs, hazard, save_mat=True)
        self.assertIsInstance(impact.imp_mat, sparse.csr_matrix)
        self.assertEqual(impact.imp_mat.shape, (hazard.event_id.size,
                                                ent.exposures.gdf.value.size))
        np.testing.assert_array_almost_equal_nulp(
            np.array(impact.imp_mat.sum(axis=1)).ravel(), impact.at_event, nulp=5)
        np.testing.assert_array_almost_equal_nulp(
            np.sum(impact.imp_mat.toarray() * impact.frequency[:, None], axis=0).reshape(-1),
            impact.eai_exp)

    def test_calc_impf_pass(self):
        """Execute when no impf_HAZ present, but only impf_"""
        ent = Entity.from_excel(ENT_DEMO_TODAY)
        self.assertTrue('impf_TC' in ent.exposures.gdf.columns)
        ent.exposures.gdf.rename(columns={'impf_TC': 'impf_'}, inplace=True)
        self.assertFalse('impf_TC' in ent.exposures.gdf.columns)
        ent.check()

        # Read default hazard file
        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        # Create impact object
        impact = Impact()
        impact.calc(ent.exposures, ent.impact_funcs, hazard)

        # Check result
        num_events = len(hazard.event_id)
        num_exp = ent.exposures.gdf.shape[0]
        # Check relative errors as well when absolute value gt 1.0e-7
        # impact.at_event == EDS.damage in MATLAB
        self.assertEqual(num_events, len(impact.at_event))
        self.assertEqual(0, impact.at_event[0])
        self.assertEqual(0, impact.at_event[int(num_events / 2)])
        self.assertAlmostEqual(1.472482938320243e+08, impact.at_event[13809])
        self.assertEqual(7.076504723057620e+10, impact.at_event[12147])
        self.assertEqual(0, impact.at_event[num_events - 1])
        # impact.eai_exp == EDS.ED_at_centroid in MATLAB
        self.assertEqual(num_exp, len(impact.eai_exp))
        self.assertAlmostEqual(1.518553670803242e+08, impact.eai_exp[0])
        self.assertAlmostEqual(1.373490457046383e+08, impact.eai_exp[int(num_exp / 2)], 6)
        self.assertAlmostEqual(1.373490457046383e+08, impact.eai_exp[int(num_exp / 2)], 5)
        self.assertAlmostEqual(1.066837260150042e+08, impact.eai_exp[num_exp - 1], 6)
        self.assertAlmostEqual(1.066837260150042e+08, impact.eai_exp[int(num_exp - 1)], 5)
        # impact.tot_value == EDS.Value in MATLAB
        # impact.aai_agg == EDS.ED in MATLAB
        self.assertAlmostEqual(6.570532945599105e+11, impact.tot_value)
        self.assertAlmostEqual(6.512201157564421e+09, impact.aai_agg, 5)
        self.assertAlmostEqual(6.512201157564421e+09, impact.aai_agg, 5)

class TestImpactYearSet(unittest.TestCase):
    """Test calc_impact_year_set method"""

    def test_impact_year_set_sum(self):
        """Test result against reference value with given events"""
        imp = Impact()
        imp.frequency = np.ones(10) * 6.211180124223603e-04
        imp.at_event = np.zeros(10)
        imp.at_event[0] = 0
        imp.at_event[1] = 0.400665463736549e9
        imp.at_event[2] = 3.150330960044466e9
        imp.at_event[3] = 3.715826406781887e9
        imp.at_event[4] = 2.900244271902339e9
        imp.at_event[5] = 0.778570745161971e9
        imp.at_event[6] = 0.698736262566472e9
        imp.at_event[7] = 0.381063674256423e9
        imp.at_event[8] = 0.569142464157450e9
        imp.at_event[9] = 0.467572545849132e9
        imp.unit = 'USD'
        imp.date = np.array([732801, 716160, 718313, 712468, 732802,
                             729285, 732931, 715419, 722404, 718351])

        iys_all = imp.calc_impact_year_set()
        iys = imp.calc_impact_year_set(all_years=False)
        iys_all_yr = imp.calc_impact_year_set(year_range=(1975, 2000))
        iys_yr = imp.calc_impact_year_set(all_years=False, year_range=[1975, 2000])
        iys_all_yr_1940 = imp.calc_impact_year_set(all_years=True, year_range=[1940, 2000])
        self.assertEqual(np.around(sum([iys[year] for year in iys])),
                         np.around(sum(imp.at_event)))
        self.assertEqual(sum([iys[year] for year in iys]),
                         sum([iys_all[year] for year in iys_all]))
        self.assertEqual(len(iys), 7)
        self.assertEqual(len(iys_all), 57)
        self.assertIn(1951 and 1959 and 2007, iys_all)
        self.assertTrue(iys_all[1959] > 0)
        self.assertAlmostEqual(3598980534.468811, iys_all[2007])
        self.assertEqual(iys[1978], iys_all[1978])
        self.assertAlmostEqual(iys[1951], imp.at_event[3])
        # year range (yr):
        self.assertEqual(len(iys_yr), 2)
        self.assertEqual(len(iys_all_yr), 26)
        self.assertEqual(sum([iys_yr[year] for year in iys_yr]),
                         sum([iys_all_yr[year] for year in iys_all_yr]))
        self.assertIn(1997 and 1978, iys_yr)
        self.assertFalse(2007 in iys_yr)
        self.assertFalse(1959 in iys_yr)
        self.assertEqual(len(iys_all_yr_1940), 61)

    def test_impact_year_set_empty(self):
        """Test result for empty impact"""
        imp = Impact()
        iys_all = imp.calc_impact_year_set()
        iys = imp.calc_impact_year_set(all_years=False)
        self.assertEqual(len(iys), 0)
        self.assertEqual(len(iys_all), 0)

class TestIO(unittest.TestCase):
    """Test impact input/output methods."""

    def test_write_read_ev_test(self):
        """Test result against reference value"""
        # Create impact object
        num_ev = 10
        num_exp = 5
        imp_write = Impact()
        imp_write.tag = {'exp': Tag('file_exp.p', 'descr exp'),
                         'haz': TagHaz('TC', 'file_haz.p', 'descr haz'),
                         'impf_set': Tag()}
        imp_write.event_id = np.arange(num_ev)
        imp_write.event_name = ['event_' + str(num) for num in imp_write.event_id]
        imp_write.date = np.ones(num_ev)
        imp_write.coord_exp = np.zeros((num_exp, 2))
        imp_write.coord_exp[:, 0] = 1.5
        imp_write.coord_exp[:, 1] = 2.5
        imp_write.eai_exp = np.arange(num_exp) * 100
        imp_write.at_event = np.arange(num_ev) * 50
        imp_write.frequency = np.ones(num_ev) * 0.1
        imp_write.tot_value = 1000
        imp_write.aai_agg = 1001
        imp_write.unit = 'USD'

        file_name = DATA_FOLDER.joinpath('test.csv')
        imp_write.write_csv(file_name)

        imp_read = Impact.from_csv(file_name)
        np.testing.assert_array_equal(imp_write.event_id, imp_read.event_id)
        np.testing.assert_array_equal(imp_write.date, imp_read.date)
        np.testing.assert_array_equal(imp_write.coord_exp, imp_read.coord_exp)
        np.testing.assert_array_equal(imp_write.eai_exp, imp_read.eai_exp)
        np.testing.assert_array_equal(imp_write.at_event, imp_read.at_event)
        np.testing.assert_array_equal(imp_write.frequency, imp_read.frequency)
        self.assertEqual(imp_write.tot_value, imp_read.tot_value)
        self.assertEqual(imp_write.aai_agg, imp_read.aai_agg)
        self.assertEqual(imp_write.unit, imp_read.unit)
        self.assertEqual(
            0, len([i for i, j in zip(imp_write.event_name, imp_read.event_name) if i != j]))

    def test_write_read_exp_test(self):
        """Test result against reference value"""
        # Create impact object
        num_ev = 5
        num_exp = 10
        imp_write = Impact()
        imp_write.tag = {'exp': Tag('file_exp.p', 'descr exp'),
                         'haz': TagHaz('TC', 'file_haz.p', 'descr haz'),
                         'impf_set': Tag()}
        imp_write.event_id = np.arange(num_ev)
        imp_write.event_name = ['event_' + str(num) for num in imp_write.event_id]
        imp_write.date = np.ones(num_ev)
        imp_write.coord_exp = np.zeros((num_exp, 2))
        imp_write.coord_exp[:, 0] = 1.5
        imp_write.coord_exp[:, 1] = 2.5
        imp_write.eai_exp = np.arange(num_exp) * 100
        imp_write.at_event = np.arange(num_ev) * 50
        imp_write.frequency = np.ones(num_ev) * 0.1
        imp_write.tot_value = 1000
        imp_write.aai_agg = 1001
        imp_write.unit = 'USD'

        file_name = DATA_FOLDER.joinpath('test.csv')
        imp_write.write_csv(file_name)

        imp_read = Impact.from_csv(file_name)
        np.testing.assert_array_equal(imp_write.event_id, imp_read.event_id)
        np.testing.assert_array_equal(imp_write.date, imp_read.date)
        np.testing.assert_array_equal(imp_write.coord_exp, imp_read.coord_exp)
        np.testing.assert_array_equal(imp_write.eai_exp, imp_read.eai_exp)
        np.testing.assert_array_equal(imp_write.at_event, imp_read.at_event)
        np.testing.assert_array_equal(imp_write.frequency, imp_read.frequency)
        self.assertEqual(imp_write.tot_value, imp_read.tot_value)
        self.assertEqual(imp_write.aai_agg, imp_read.aai_agg)
        self.assertEqual(imp_write.unit, imp_read.unit)
        self.assertEqual(
            0, len([i for i, j in zip(imp_write.event_name, imp_read.event_name) if i != j]))
        self.assertIsInstance(imp_read.crs, str)

    def test_excel_io(self):
        """Test write and read in excel"""
        ent = Entity.from_excel(ENT_DEMO_TODAY)
        ent.check()

        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        imp_write = Impact()
        ent.exposures.assign_centroids(hazard)
        imp_write.calc(ent.exposures, ent.impact_funcs, hazard)
        file_name = DATA_FOLDER.joinpath('test.xlsx')
        imp_write.write_excel(file_name)

        imp_read = Impact.from_excel(file_name)

        np.testing.assert_array_equal(imp_write.event_id, imp_read.event_id)
        np.testing.assert_array_equal(imp_write.date, imp_read.date)
        np.testing.assert_array_equal(imp_write.coord_exp, imp_read.coord_exp)
        np.testing.assert_array_almost_equal_nulp(imp_write.eai_exp, imp_read.eai_exp, nulp=5)
        np.testing.assert_array_almost_equal_nulp(imp_write.at_event, imp_read.at_event, nulp=5)
        np.testing.assert_array_equal(imp_write.frequency, imp_read.frequency)
        self.assertEqual(imp_write.tot_value, imp_read.tot_value)
        self.assertEqual(imp_write.aai_agg, imp_read.aai_agg)
        self.assertEqual(imp_write.unit, imp_read.unit)
        self.assertEqual(
            0, len([i for i, j in zip(imp_write.event_name, imp_read.event_name) if i != j]))
        self.assertIsInstance(imp_read.crs, str)

    def test_write_imp_mat(self):
        """Test write_excel_imp_mat function"""
        impact = Impact()
        impact.imp_mat = np.zeros((5, 4))
        impact.imp_mat[0, :] = np.arange(4)
        impact.imp_mat[1, :] = np.arange(4) * 2
        impact.imp_mat[2, :] = np.arange(4) * 3
        impact.imp_mat[3, :] = np.arange(4) * 4
        impact.imp_mat[4, :] = np.arange(4) * 5
        impact.imp_mat = sparse.csr_matrix(impact.imp_mat)

        file_name = DATA_FOLDER.joinpath('test_imp_mat')
        impact.write_sparse_csr(file_name)
        read_imp_mat = Impact().read_sparse_csr(f'{file_name}.npz')
        for irow in range(5):
            np.testing.assert_array_equal(
                read_imp_mat[irow, :].toarray(), impact.imp_mat[irow, :].toarray())

class TestRPmatrix(unittest.TestCase):
    """Test computation of impact per return period for whole exposure"""
    def test_local_exceedance_imp_pass(self):
        """Test calc local impacts per return period"""
        # Read default entity values
        ent = Entity.from_excel(ENT_DEMO_TODAY)
        ent.check()

        # Read default hazard file
        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        # Create impact object
        impact = Impact()
        # Assign centroids to exposures
        ent.exposures.assign_centroids(hazard)
        # Compute the impact over the whole exposures
        impact.calc(ent.exposures, ent.impact_funcs, hazard, save_mat=True)
        # Compute the impact per return period over the whole exposures
        impact_rp = impact.local_exceedance_imp(return_periods=(10, 40))

        self.assertIsInstance(impact_rp, np.ndarray)
        self.assertEqual(impact_rp.size, 2 * ent.exposures.gdf.value.size)
        self.assertAlmostEqual(np.max(impact_rp), 2916964966.388219, places=5)
        self.assertAlmostEqual(np.min(impact_rp), 444457580.131494, places=5)

class TestRiskTrans(unittest.TestCase):
    """Test risk transfer methods"""
    def test_risk_trans_pass(self):
        """Test calc_risk_transfer"""
        # Create impact object
        imp = Impact()
        imp.event_id = np.arange(10)
        imp.event_name = [0, 1, 2, 3, 4, 5, 6, 7, 8, 15]
        imp.date = np.arange(10)
        imp.coord_exp = np.array([[1, 2], [2, 3]])
        imp.crs = DEF_CRS
        imp.eai_exp = np.array([1, 2])
        imp.at_event = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 15])
        imp.frequency = np.ones(10) / 5
        imp.tot_value = 10
        imp.aai_agg = 100
        imp.unit = 'USD'
        imp.imp_mat = sparse.csr_matrix(np.empty((0, 0)))

        new_imp, imp_rt = imp.calc_risk_transfer(2, 10)
        self.assertEqual(new_imp.unit, imp.unit)
        self.assertEqual(new_imp.tot_value, imp.tot_value)
        np.testing.assert_array_equal(new_imp.imp_mat.toarray(), imp.imp_mat.toarray())
        self.assertEqual(new_imp.event_name, imp.event_name)
        np.testing.assert_array_almost_equal_nulp(new_imp.event_id, imp.event_id)
        np.testing.assert_array_almost_equal_nulp(new_imp.date, imp.date)
        np.testing.assert_array_almost_equal_nulp(new_imp.frequency, imp.frequency)
        np.testing.assert_array_almost_equal_nulp(new_imp.coord_exp, [])
        np.testing.assert_array_almost_equal_nulp(new_imp.eai_exp, [])
        np.testing.assert_array_almost_equal_nulp(new_imp.at_event, [0, 1, 2, 2, 2, 2, 2, 2, 2, 5])
        self.assertAlmostEqual(new_imp.aai_agg, 4.0)

        self.assertEqual(imp_rt.unit, imp.unit)
        self.assertEqual(imp_rt.tot_value, imp.tot_value)
        np.testing.assert_array_equal(imp_rt.imp_mat.toarray(), imp.imp_mat.toarray())
        self.assertEqual(imp_rt.event_name, imp.event_name)
        np.testing.assert_array_almost_equal_nulp(imp_rt.event_id, imp.event_id)
        np.testing.assert_array_almost_equal_nulp(imp_rt.date, imp.date)
        np.testing.assert_array_almost_equal_nulp(imp_rt.frequency, imp.frequency)
        np.testing.assert_array_almost_equal_nulp(imp_rt.coord_exp, [])
        np.testing.assert_array_almost_equal_nulp(imp_rt.eai_exp, [])
        np.testing.assert_array_almost_equal_nulp(imp_rt.at_event, [0, 0, 0, 1, 2, 3, 4, 5, 6, 10])
        self.assertAlmostEqual(imp_rt.aai_agg, 6.2)

def dummy_impact():

    imp = Impact()
    imp.event_id = np.arange(6) + 10
    imp.event_name = [0, 1, 'two', 'three', 30, 31]
    imp.date = np.arange(6)
    imp.coord_exp = np.array([[1, 2], [1.5, 2.5]])
    imp.crs = DEF_CRS
    imp.eai_exp = np.array([7.2, 7.2])
    imp.at_event = np.array([0, 2, 4, 6, 60, 62])
    imp.frequency = np.array([1/6, 1/6, 1, 1, 1/30, 1/30])
    imp.tot_value = 7
    imp.aai_agg = 14.4
    imp.unit = 'USD'
    imp.imp_mat = sparse.csr_matrix(np.array([
        [0,0], [1,1], [2,2], [3,3], [30,30], [31,31]
        ]))

    return imp

class TestSelect(unittest.TestCase):
    """Test select method"""
    def test_select_event_id_pass(self):
        """Test select by event id"""

        imp = dummy_impact()
        sel_imp = imp.select(event_ids=[10, 11, 12])

        self.assertTrue(u_coord.equal_crs(sel_imp.crs, imp.crs))
        self.assertEqual(sel_imp.unit, imp.unit)

        np.testing.assert_array_equal(sel_imp.event_id, [10, 11, 12])
        self.assertEqual(sel_imp.event_name, [0, 1, 'two'])
        np.testing.assert_array_equal(sel_imp.date, [0, 1, 2])
        np.testing.assert_array_almost_equal_nulp(sel_imp.frequency, [1/6, 1/6, 1])

        np.testing.assert_array_equal(sel_imp.at_event, [0, 2, 4])
        np.testing.assert_array_equal(sel_imp.imp_mat.todense(), [[0,0], [1,1], [2,2]])
        np.testing.assert_array_almost_equal_nulp(sel_imp.eai_exp, [1/6+2, 1/6+2])
        self.assertEqual(sel_imp.aai_agg, 4+2/6)

        self.assertEqual(sel_imp.tot_value, 7)
        np.testing.assert_array_equal(sel_imp.coord_exp, [[1, 2], [1.5, 2.5]])

        self.assertIsInstance(sel_imp, Impact)
        self.assertIsInstance(sel_imp.imp_mat, sparse.csr_matrix)

    def test_select_event_name_pass(self):
        """Test select by event name"""

        imp = dummy_impact()
        sel_imp = imp.select(event_names=[0, 1, 'two'])

        self.assertTrue(u_coord.equal_crs(sel_imp.crs, imp.crs))
        self.assertEqual(sel_imp.unit, imp.unit)

        np.testing.assert_array_equal(sel_imp.event_id, [10, 11, 12])
        self.assertEqual(sel_imp.event_name, [0, 1, 'two'])
        np.testing.assert_array_equal(sel_imp.date, [0, 1, 2])
        np.testing.assert_array_almost_equal_nulp(sel_imp.frequency, [1/6, 1/6, 1])

        np.testing.assert_array_equal(sel_imp.at_event, [0, 2, 4])
        np.testing.assert_array_equal(sel_imp.imp_mat.todense(), [[0,0], [1,1], [2,2]])
        np.testing.assert_array_almost_equal_nulp(sel_imp.eai_exp, [1/6+2, 1/6+2])
        self.assertEqual(sel_imp.aai_agg, 4+2/6)

        self.assertEqual(sel_imp.tot_value, 7)
        np.testing.assert_array_equal(sel_imp.coord_exp, [[1, 2], [1.5, 2.5]])

        self.assertIsInstance(sel_imp, Impact)
        self.assertIsInstance(sel_imp.imp_mat, sparse.csr_matrix)

    def test_select_dates_pass(self):
        """Test select by event dates"""

        imp = dummy_impact()
        sel_imp = imp.select(dates=(0, 2))

        self.assertTrue(u_coord.equal_crs(sel_imp.crs, imp.crs))
        self.assertEqual(sel_imp.unit, imp.unit)

        np.testing.assert_array_equal(sel_imp.event_id, [10, 11, 12])
        self.assertEqual(sel_imp.event_name, [0, 1, 'two'])
        np.testing.assert_array_equal(sel_imp.date, [0, 1, 2])
        np.testing.assert_array_almost_equal_nulp(sel_imp.frequency, [1/6, 1/6, 1])

        np.testing.assert_array_equal(sel_imp.at_event, [0, 2, 4])
        np.testing.assert_array_equal(sel_imp.imp_mat.todense(), [[0,0], [1,1], [2,2]])
        np.testing.assert_array_almost_equal_nulp(sel_imp.eai_exp, [1/6+2, 1/6+2])
        self.assertEqual(sel_imp.aai_agg, 4+2/6)

        self.assertEqual(sel_imp.tot_value, 7)
        np.testing.assert_array_equal(sel_imp.coord_exp, [[1, 2], [1.5, 2.5]])

        self.assertIsInstance(sel_imp, Impact)
        self.assertIsInstance(sel_imp.imp_mat, sparse.csr_matrix)

    def test_select_coord_exp_pass(self):
        """ test select by exp coordinates """

        imp = dummy_impact()
        sel_imp = imp.select(coord_exp=np.array([1,2]))

        self.assertTrue(u_coord.equal_crs(sel_imp.crs, imp.crs))
        self.assertEqual(sel_imp.unit, imp.unit)

        np.testing.assert_array_equal(sel_imp.event_id, imp.event_id)
        self.assertEqual(sel_imp.event_name, imp.event_name)
        np.testing.assert_array_equal(sel_imp.date, imp.date)
        np.testing.assert_array_equal(sel_imp.frequency, imp.frequency)

        np.testing.assert_array_equal(sel_imp.at_event, [0, 1, 2, 3, 30, 31])
        np.testing.assert_array_equal(sel_imp.imp_mat.todense(), [[0], [1], [2], [3], [30], [31]])
        np.testing.assert_array_almost_equal_nulp(sel_imp.eai_exp, [1/6+2+3+1+31/30])
        self.assertEqual(sel_imp.aai_agg, 1/6+2+3+1+31/30)

        self.assertEqual(sel_imp.tot_value, None)
        np.testing.assert_array_equal(sel_imp.coord_exp, [[1, 2]])

        self.assertIsInstance(sel_imp, Impact)
        self.assertIsInstance(sel_imp.imp_mat, sparse.csr_matrix)

    def test_select_event_identity_pass(self):
        """ test select same impact with event name, id and date """

        # Read default entity values
        ent = Entity.from_excel(ENT_DEMO_TODAY)
        ent.check()

        # Read default hazard file
        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        # Create impact object
        imp = Impact()

        # Assign centroids to exposures
        ent.exposures.assign_centroids(hazard)

        # Compute the impact over the whole exposures
        imp.calc(ent.exposures, ent.impact_funcs, hazard, save_mat=True)

        sel_imp = imp.select(event_ids=imp.event_id,
                             event_names=imp.event_name,
                             dates=(min(imp.date), max(imp.date))
                             )

        self.assertTrue(u_coord.equal_crs(sel_imp.crs, imp.crs))
        self.assertEqual(sel_imp.unit, imp.unit)

        np.testing.assert_array_equal(sel_imp.event_id, imp.event_id)
        self.assertEqual(sel_imp.event_name, imp.event_name)
        np.testing.assert_array_equal(sel_imp.date, imp.date)
        np.testing.assert_array_equal(sel_imp.frequency, imp.frequency)

        np.testing.assert_array_equal(sel_imp.at_event, imp.at_event)
        np.testing.assert_array_equal(sel_imp.imp_mat.todense(), imp.imp_mat.todense())
        np.testing.assert_array_equal(sel_imp.eai_exp, imp.eai_exp)
        self.assertAlmostEqual(round(sel_imp.aai_agg,5), round(imp.aai_agg,5))

        self.assertEqual(sel_imp.tot_value, imp.tot_value)
        np.testing.assert_array_equal(sel_imp.coord_exp, imp.coord_exp)

        self.assertIsInstance(sel_imp, Impact)
        self.assertIsInstance(sel_imp.imp_mat, sparse.csr_matrix)


    def test_select_new_attributes(self):
        """Test if impact has new attributes """

        imp = dummy_impact()
        imp.new_per_ev =  ['a', 'b', 'c', 'd', 'e', 'f']
        sel_imp = imp.select(event_names=[0, 1, 'two'])

        self.assertEqual(sel_imp.new_per_ev, ['a', 'b', 'c'])

        self.assertTrue(u_coord.equal_crs(sel_imp.crs, imp.crs))
        self.assertEqual(sel_imp.unit, imp.unit)

        np.testing.assert_array_equal(sel_imp.event_id, [10, 11, 12])
        self.assertEqual(sel_imp.event_name, [0, 1, 'two'])
        np.testing.assert_array_equal(sel_imp.date, [0, 1, 2])
        np.testing.assert_array_almost_equal_nulp(sel_imp.frequency, [1/6, 1/6, 1])

        np.testing.assert_array_equal(sel_imp.at_event, [0, 2, 4])
        np.testing.assert_array_equal(sel_imp.imp_mat.todense(), [[0,0], [1,1], [2,2]])
        np.testing.assert_array_almost_equal_nulp(sel_imp.eai_exp, [1/6+2, 1/6+2])
        self.assertEqual(sel_imp.aai_agg, 4+2/6)

        self.assertEqual(sel_imp.tot_value, 7)
        np.testing.assert_array_equal(sel_imp.coord_exp, [[1, 2], [1.5, 2.5]])

        self.assertIsInstance(sel_imp, Impact)
        self.assertIsInstance(sel_imp.imp_mat, sparse.csr_matrix)

    def test_select_nothing(self):
        """Test select with no matches"""
        imp = dummy_impact()
        sel_imp = imp.select(event_ids=[100])
        self.assertIsInstance(sel_imp, Impact)
        self.assertIsInstance(sel_imp.imp_mat, sparse.csr_matrix)
        self.assertTrue(u_coord.equal_crs(sel_imp.crs, imp.crs))
        self.assertEqual(sel_imp.unit, imp.unit)
        self.assertEqual(sel_imp.event_id.size, 0)
        self.assertEqual(len(sel_imp.event_name), 0)
        self.assertEqual(sel_imp.date.size, 0)
        self.assertEqual(sel_imp.frequency.size, 0)
        self.assertEqual(sel_imp.at_event.size, 0)
        self.assertEqual(sel_imp.imp_mat.shape[0], 0)
        self.assertEqual(sel_imp.aai_agg, 0)

    def test_select_id_name_dates_pass(self):
        """Test select by event ids, names, and dates"""

        imp = dummy_impact()
        sel_imp = imp.select(event_ids=[0], event_names=[1, 'two'], dates=(0, 2))

        self.assertTrue(u_coord.equal_crs(sel_imp.crs, imp.crs))
        self.assertEqual(sel_imp.unit, imp.unit)

        np.testing.assert_array_equal(sel_imp.event_id, [10, 11, 12])
        self.assertEqual(sel_imp.event_name, [0, 1, 'two'])
        np.testing.assert_array_equal(sel_imp.date, [0, 1, 2])
        np.testing.assert_array_almost_equal_nulp(sel_imp.frequency, [1/6, 1/6, 1])

        np.testing.assert_array_equal(sel_imp.at_event, [0, 2, 4])
        np.testing.assert_array_equal(sel_imp.imp_mat.todense(), [[0,0], [1,1], [2,2]])
        np.testing.assert_array_almost_equal_nulp(sel_imp.eai_exp, [1/6+2, 1/6+2])
        self.assertEqual(sel_imp.aai_agg, 4+2/6)

        self.assertEqual(sel_imp.tot_value, 7)
        np.testing.assert_array_equal(sel_imp.coord_exp, [[1, 2], [1.5, 2.5]])

        self.assertIsInstance(sel_imp, Impact)
        self.assertIsInstance(sel_imp.imp_mat, sparse.csr_matrix)

    def test_select_imp_map_fail(self):
        """Test that selection fails if imp_mat is empty"""

        imp = dummy_impact()
        imp.imp_mat = sparse.csr_matrix(np.empty((0, 0)))
        with self.assertRaises(ValueError):
            imp.select(event_ids=[0], event_names=[1, 'two'], dates=(0, 2))

    def test__build_exp(self):
        """Test that an impact set can be converted to an exposure"""

        imp = dummy_impact()
        exp = imp._build_exp()
        np.testing.assert_array_equal(imp.eai_exp, exp.gdf['value'])
        np.testing.assert_array_equal(imp.coord_exp[:, 0], exp.gdf['latitude'])
        np.testing.assert_array_equal(imp.coord_exp[:, 1], exp.gdf['longitude'])
        self.assertTrue(u_coord.equal_crs(exp.crs, imp.crs))
        self.assertEqual(exp.value_unit, imp.unit)
        self.assertEqual(exp.ref_year, 0)

    def test__exp_build_event(self):
        """Test that a single events impact can be converted to an exposure"""

        imp = dummy_impact()
        event_id = imp.event_id[1]
        exp = imp._build_exp_event(event_id=event_id)
        np.testing.assert_array_equal(imp.imp_mat[1].todense().A1, exp.gdf['value'])
        np.testing.assert_array_equal(imp.coord_exp[:, 0], exp.gdf['latitude'])
        np.testing.assert_array_equal(imp.coord_exp[:, 1], exp.gdf['longitude'])
        self.assertTrue(u_coord.equal_crs(exp.crs, imp.crs))
        self.assertEqual(exp.value_unit, imp.unit)
        self.assertEqual(exp.ref_year, 0)


# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestOneExposure)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestCalc))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestFreqCurve))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestImpactYearSet))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestIO))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestRPmatrix))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestRiskTrans))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSelect))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define Uncertainty Cost Benefit class
"""

__all__ = ['CalcCostBenefit']

import logging
import time

from functools import partial
import pandas as pd

from climada.engine.cost_benefit import CostBenefit
from climada.engine.unsequa import Calc, InputVar, UncCostBenefitOutput
from climada.util import log_level

LOGGER = logging.getLogger(__name__)

# Future planed features:
# - Add 'efc' (frequency curve) to UncCostBenenfit

class CalcCostBenefit(Calc):
    """
    Cost Benefit uncertainty analysis class

    This is the base class to perform uncertainty analysis on the outputs of
    climada.engine.costbenefit.CostBenefit().

    Attributes
    ----------
    metric_names : tuple(str)
        Names of the cost benefit output metris
        ('tot_climate_risk', 'benefit', 'cost_ben_ratio',
         'imp_meas_present', 'imp_meas_future')
    value_unit : str
        Unit of the exposures value
    input_var_names : tuple(str)
        Names of the required uncertainty variables
        ('haz_input_var', 'ent_input_var', 'haz_fut_input_var', 'ent_fut_input_var')
    haz_input_var : climada.engine.uncertainty.input_var.InputVar
        Present Hazard uncertainty variable
    ent_input_var : climada.engine.uncertainty.input_var.InputVar
        Present Entity uncertainty variable
    haz_unc_fut_Var: climada.engine.uncertainty.input_var.InputVar
        Future Hazard uncertainty variable
    ent_fut_input_var : climada.engine.uncertainty.input_var.InputVar
        Future Entity uncertainty variable

    """

    def __init__(self, haz_input_var, ent_input_var,
                 haz_fut_input_var=None, ent_fut_input_var=None):
        """Initialize UncCalcCostBenefit

        Sets the uncertainty input variables, the cost benefit metric_names,
        and the units.

        Parameters
        ----------
        haz_input_var : climada.engine.uncertainty.input_var.InputVar
                        or climada.hazard.Hazard
            Hazard uncertainty variable or Hazard for the present Hazard
            in climada.engine.CostBenefit.calc
        ent_input_var : climada.engine.uncertainty.input_var.InputVar
                        or climada.entity.Entity
            Entity uncertainty variable or Entity for the present Entity
            in climada.engine.CostBenefit.calc
        haz_fut_input_var: climada.engine.uncertainty.input_var.InputVar
                           or climada.hazard.Hazard, optional
            Hazard uncertainty variable or Hazard for the future Hazard
            The Default is None.
        ent_fut_input_var : climada.engine.uncertainty.input_var.InputVar
                            or climada.entity.Entity, optional
            Entity uncertainty variable or Entity for the future Entity
            in climada.engine.CostBenefit.calc

        """

        Calc.__init__(self)
        self.input_var_names = ('haz_input_var', 'ent_input_var',
                             'haz_fut_input_var', 'ent_fut_input_var')
        self.haz_input_var = InputVar.var_to_inputvar(haz_input_var)
        self.ent_input_var = InputVar.var_to_inputvar(ent_input_var)
        self.haz_fut_input_var = InputVar.var_to_inputvar(haz_fut_input_var)
        self.ent_fut_input_var = InputVar.var_to_inputvar(ent_fut_input_var)
        self.metric_names = ('tot_climate_risk', 'benefit',
                             'cost_ben_ratio',
                             'imp_meas_present', 'imp_meas_future')
        self.value_unit = self.ent_input_var.evaluate().exposures.value_unit
        self.check_distr()



    def uncertainty(self, unc_data, pool=None, **cost_benefit_kwargs):
        """
        Computes the cost benefit for each sample in unc_output.sample_df.

        By default, imp_meas_present, imp_meas_future, tot_climate_risk,
        benefit, cost_ben_ratio are computed.

        This sets the attributes:
        unc_output.imp_meas_present_unc_df,
        unc_output.imp_meas_future_unc_df
        unc_output.tot_climate_risk_unc_df
        unc_output.benefit_unc_df
        unc_output.cost_ben_ratio_unc_df
        unc_output.unit
        unc_output.cost_benefit_kwargs

        Parameters
        ----------
        unc_data : climada.engine.uncertainty.unc_output.UncOutput
            Uncertainty data object with the input parameters samples
        pool : pathos.pools.ProcessPool, optional
            Pool of CPUs for parralel computations. Default is None.
            The default is None.
        cost_benefit_kwargs : keyword arguments
            Keyword arguments passed on to climada.engine.CostBenefit.calc()

        Returns
        -------
        unc_output : climada.engine.uncertainty.unc_output.UncCostBenefitOutput
            Uncertainty data object in with the cost benefit outputs for each
            sample and all the sample data copied over from unc_sample.

        Raises
        ------
        ValueError:
            If no sampling parameters defined, the uncertainty distribution
            cannot be computed.

        See Also
        --------
        climada.engine.cost_benefit:
            Compute risk and adptation option cost benefits.

        """

        if unc_data.samples_df.empty:
            raise ValueError("No sample was found. Please create one first" +
                        "using UncImpact.make_sample(N)")

        samples_df = unc_data.samples_df.copy(deep=True)
        unit = self.value_unit

        LOGGER.info("The freq_curve is not saved. Please "
                    "change the risk_func (see climada.engine.cost_benefit) "
                    "if return period information is needed")

        start = time.time()
        one_sample = samples_df.iloc[0:1].iterrows()
        cb_metrics = map(self._map_costben_calc, one_sample)
        [imp_meas_present,
         imp_meas_future,
         tot_climate_risk,
         benefit,
         cost_ben_ratio] = list(zip(*cb_metrics))
        elapsed_time = (time.time() - start)
        self.est_comp_time(unc_data.n_samples, elapsed_time, pool)

        #Compute impact distributions
        with log_level(level='ERROR', name_prefix='climada'):
            if pool:
                LOGGER.info('Using %s CPUs.', pool.ncpus)
                chunksize = min(unc_data.n_samples // pool.ncpus, 100)
                cb_metrics = pool.map(partial(self._map_costben_calc, **cost_benefit_kwargs),
                                               samples_df.iterrows(),
                                               chunsize = chunksize)

            else:
                cb_metrics = map(partial(self._map_costben_calc, **cost_benefit_kwargs),
                                 samples_df.iterrows())

        #Perform the actual computation
        with log_level(level='ERROR', name_prefix='climada'):
            [imp_meas_present,
             imp_meas_future,
             tot_climate_risk,
             benefit,
             cost_ben_ratio] = list(zip(*cb_metrics)) #Transpose list of list

        # Assign computed impact distribution data to self
        tot_climate_risk_unc_df = \
            pd.DataFrame(tot_climate_risk, columns = ['tot_climate_risk'])

        benefit_unc_df = pd.DataFrame(benefit)
        benefit_unc_df.columns = [
            column + ' Benef'
            for column in benefit_unc_df.columns]
        cost_ben_ratio_unc_df = pd.DataFrame(cost_ben_ratio)
        cost_ben_ratio_unc_df.columns = [
            column + ' CostBen'
            for column in cost_ben_ratio_unc_df.columns]

        imp_metric_names = ['risk', 'risk_transf', 'cost_meas',
                            'cost_ins']

        im_periods = dict()
        for imp_meas, period in zip([imp_meas_present, imp_meas_future],
                                  ['present', 'future']):
            df_imp_meas = pd.DataFrame()
            name = 'imp_meas_' + period
            if imp_meas[0]:
                for imp in imp_meas:
                    met_dic = {}
                    for meas, imp_dic in imp.items():
                        metrics = [imp_dic['risk'],
                                   imp_dic['risk_transf'],
                                   *imp_dic['cost']]
                        dic_tmp = {meas + ' - ' + m_name + ' - ' + period: [m_value]
                                   for m_name, m_value
                                   in zip(imp_metric_names, metrics)
                                    }
                        met_dic.update(dic_tmp)
                    df_imp_meas = df_imp_meas.append(
                        pd.DataFrame(met_dic), ignore_index=True
                        )
            im_periods[name + '_unc_df'] = df_imp_meas
        cost_benefit_kwargs = {
            key: str(val)
            for key, val in cost_benefit_kwargs.items()}
        cost_benefit_kwargs = tuple(cost_benefit_kwargs.items())

        return UncCostBenefitOutput(samples_df=samples_df,
                                    imp_meas_present_unc_df=im_periods['imp_meas_present_unc_df'],
                                    imp_meas_future_unc_df=im_periods['imp_meas_future_unc_df'],
                                    tot_climate_risk_unc_df=tot_climate_risk_unc_df,
                                    cost_ben_ratio_unc_df=cost_ben_ratio_unc_df,
                                    benefit_unc_df=benefit_unc_df,
                                    unit=unit,
                                    cost_benefit_kwargs=cost_benefit_kwargs)

    def _map_costben_calc(self, param_sample, **kwargs):
        """
        Map to compute cost benefit for all parameter samples in parallel

        Parameters
        ----------
        param_sample : pd.DataFrame.iterrows()
            Generator of the parameter samples
        kwargs :
            Keyword arguments passed on to climada.engine.CostBenefit.calc()

        Returns
        -------
        list
            icost benefit metrics list for all samples containing
            imp_meas_present, imp_meas_future, tot_climate_risk,
            benefit, cost_ben_ratio

        """

        # [1] only the rows of the dataframe passed by pd.DataFrame.iterrows()
        haz_samples = param_sample[1][self.haz_input_var.labels].to_dict()
        ent_samples = param_sample[1][self.ent_input_var.labels].to_dict()
        haz_fut_samples = param_sample[1][self.haz_fut_input_var.labels].to_dict()
        ent_fut_samples = param_sample[1][self.ent_fut_input_var.labels].to_dict()

        haz = self.haz_input_var.evaluate(**haz_samples)
        ent = self.ent_input_var.evaluate(**ent_samples)
        haz_fut = self.haz_fut_input_var.evaluate(**haz_fut_samples)
        ent_fut = self.ent_fut_input_var.evaluate(**ent_fut_samples)

        cb = CostBenefit()
        cb.calc(hazard=haz, entity=ent, haz_future=haz_fut, ent_future=ent_fut,
                save_imp=False, **kwargs)

        # Extract from climada.impact the chosen metrics
        return  [cb.imp_meas_present,
                 cb.imp_meas_future,
                 cb.tot_climate_risk,
                 cb.benefit,
                 cb.cost_ben_ratio
                 ]
